# -*- coding: utf-8 -*-
# Copyright 2018-2024 the orix developers
#
# This file is part of orix.
#
# orix is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# orix is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with orix.  If not, see <http://www.gnu.org/licenses/>.

"""Stereographic plot inheriting from :class:`~matplotlib.axes.Axes` for
plotting :class:`~orix.vector.Vector3d`.
"""

from copy import deepcopy
from typing import Any, List, Optional, Tuple, Union

from matplotlib import rcParams
import matplotlib.axes as maxes
import matplotlib.collections as mcollections
import matplotlib.patches as mpatches
import matplotlib.path as mpath
import matplotlib.projections as mprojections
import matplotlib.pyplot as plt
import numpy as np

# fmt: off
# isort: off
from orix.measure import pole_density_function
from orix.plot._symmetry_marker import (
    TwoFoldMarker,
    ThreeFoldMarker,
    FourFoldMarker,
    SixFoldMarker,
)
# isort: on
# fmt: on
from orix.projections import InverseStereographicProjection, StereographicProjection
from orix.vector import FundamentalSector, Vector3d
from orix.vector.fundamental_sector import _closed_edges_in_hemisphere

# This order determines which parts are plotted over other parts
ZORDER = dict(
    text=6,
    scatter=5,
    symmetry_marker=4,
    draw_circle=3,
    border=2,
    grid=1,
    mesh=0,
)


class StereographicPlot(maxes.Axes):
    """Stereographic plot for plotting :class:`~orix.vector.Vector3d`.

    Parameters
    ----------
    *args
        Arguments passed to :class:`matplotlib.axes.Axes`.
    hemisphere
        Which hemisphere to plot vectors in, either ``"upper"``
        (default) or ``"lower"``.
    azimuth_resolution
        Resolution of azimuth grid lines in degrees. Default is 10
        degrees.
    polar_resolution
        Resolution of polar grid lines in degrees. Default is 10
        degrees.
    **kwargs
        Keyword arguments passed to :class:`matplotlib.axes.Axes`.
    """

    name = "stereographic"
    _pad_xy = 0.05

    def __init__(
        self,
        *args,
        hemisphere: str = "upper",
        azimuth_resolution: Union[int, float] = 10,
        polar_resolution: Union[int, float] = 10,
        **kwargs,
    ):
        """Create an axis for plotting :class:`~orix.vector.Vector3d`."""
        self.hemisphere = hemisphere
        self._azimuth_resolution = azimuth_resolution
        self._polar_resolution = polar_resolution

        # Custom attribute to keep track of whether grid is on or off
        self._stereographic_grid = None

        super().__init__(*args, **kwargs)

        # Set ratio of y-unit to x-unit by adjusting the physical
        # dimension of the Axes (box), and centering the anchor (C)
        self.set_aspect("equal", adjustable="box", anchor="C")
        self.clear()

    def clear(self):
        super().clear()

        self.xaxis.set_ticks_position("none")
        self.yaxis.set_ticks_position("none")
        self.xaxis.set_tick_params(label1On=False)
        self.yaxis.set_tick_params(label1On=False)

        self.set_xlim(-1 - self._pad_xy, 1 + self._pad_xy)
        self.set_ylim(-1 - self._pad_xy, 1 + self._pad_xy)

        spines = self.spines
        for spine in spines.values():
            spine.set_visible(False)

        self.add_patch(
            mpatches.Circle(
                xy=(0, 0),
                radius=1,
                facecolor="none",
                edgecolor="k",
                label="sa_circle",
                zorder=ZORDER["border"],
            )
        )

        # Don't show rectangular grid
        self.grid(False)
        self.stereographic_grid(rcParams["axes.grid"])

    def format_coord(self, x, y):
        if np.sqrt(np.sum(np.square([x, y]))) > 1:
            return ""
        else:
            azimuth, polar = self._inverse_projection.xy2spherical(x, y)
            azimuth = azimuth[0]
            polar = polar[0]
            azimuth_deg = np.rad2deg(azimuth)
            polar_deg = np.rad2deg(polar)
            return (
                "\N{GREEK SMALL LETTER PHI}={:.2f}\N{GREEK SMALL LETTER PI} "
                "({:.2f}\N{DEGREE SIGN}), "
                "\N{GREEK SMALL LETTER theta}={:.2f}\N{GREEK SMALL LETTER PI} "
                "({:.2f}\N{DEGREE SIGN})"
            ).format(azimuth / np.pi, azimuth_deg, polar / np.pi, polar_deg)

    def plot(
        self,
        *args: Union[Vector3d, Tuple[float, float], Tuple[np.ndarray, np.ndarray]],
        **kwargs,
    ):
        """Draw straight lines between vectors.

        This method overwrites :meth:`matplotlib.axes.Axes.plot`, see
        that method's docstring for parameters.

        Parameters
        ----------
        *args
            Vector(s), or azimuth and polar angles, the latter two
            passed as separate arguments (not keyword arguments).
        **kwargs
            Keyword arguments passed to
            :meth:`matplotlib.axes.Axes.plot`.

        See Also
        --------
        matplotlib.axes.Axes.plot
        """
        new_kwargs = dict(clip_on=True, linewidth=2, color="k", linestyle="-")
        x, y, _, updated_kwargs = self._prepare_to_call_inherited_method(
            args, kwargs, new_kwargs, sort=True
        )
        if x.size == 0:
            return

        super().plot(x, y, **updated_kwargs)

    def scatter(
        self,
        *args: Union[Vector3d, Tuple[float, float], Tuple[np.ndarray, np.ndarray]],
        **kwargs,
    ):
        """A scatter plot of vectors.

        This method overwrites :meth:`matplotlib.axes.Axes.scatter`, see
        that method's docstring for parameters.

        Parameters
        ----------
        *args
            Vector(s), or azimuth and polar angles, the latter two
            passed as separate arguments (not keyword arguments).
        **kwargs
            Keyword arguments passed to
            :meth:`matplotlib.axes.Axes.scatter`.

        See Also
        --------
        matplotlib.axes.Axes.scatter
        """
        new_kwargs = dict(zorder=ZORDER["scatter"], clip_on=False)
        out = self._prepare_to_call_inherited_method(args, kwargs, new_kwargs)
        x, y, visible, updated_kwargs = out
        if x.size == 0:
            return

        # Color(s)
        if "color" in updated_kwargs.keys():
            key_color = "color"
        else:
            key_color = "c"
        c = updated_kwargs.pop(key_color, "C0")
        c = _get_array_of_values(value=c, visible=visible)

        # Size(s)
        if "sizes" in updated_kwargs.keys():
            key_size = "sizes"
        else:
            key_size = "s"
        s = updated_kwargs.pop(key_size, None)
        if s is not None:
            s = _get_array_of_values(value=s, visible=visible)

        super().scatter(x, y, c=c, s=s, **updated_kwargs)

    def text(
        self,
        *args: Union[Vector3d, Tuple[float, float], Tuple[np.ndarray, np.ndarray]],
        offset: Optional[Tuple[float, float]] = None,
        **kwargs,
    ):
        """Add text to the axes.

        This method overwrites :meth:`matplotlib.axes.Axes.text`, see
        that method's docstring for parameters.

        Parameters
        ----------
        *args
            Vector(s), or azimuth and polar angles, the latter two
            passed as separate arguments (not keyword arguments).
        offset
            Tuple of offsets in stereographic coordinates (X, Y). No
            offset is applied if not given.
        **kwargs
            Keyword arguments passed to
            :meth:`matplotlib.axes.Axes.text`.

        See Also
        --------
        matplotlib.axes.Axes.text
        """
        new_kwargs = dict(va="bottom", ha="center", zorder=ZORDER["text"])
        out = self._prepare_to_call_inherited_method(
            args, kwargs, new_kwargs, offset=offset
        )
        x, y, _, updated_kwargs = out
        if x.size == 0:
            return
        super().text(x, y, **updated_kwargs)

    # ----------- Custom attributes and methods below here ----------- #

    @property
    def hemisphere(self) -> str:
        """Return or set the hemisphere to plot, either ``"upper"`` or
        ``"lower"``.

        :attr:`pole` is derived from this attribute.

        Parameters
        ----------
        value : str
            Either ``"upper"`` or ``"lower"``.
        """
        return self._hemisphere

    @hemisphere.setter
    def hemisphere(self, value: str):
        """Set hemisphere to plot."""
        value = value.lower()
        if value in ["upper", "lower"]:
            self._hemisphere = value
        else:
            raise ValueError(f"Hemisphere must be 'upper' or 'lower', not {value}.")

    @property
    def pole(self) -> int:
        """Return the projection pole, either -1 or 1, where -1 (1)
        means the projection point of the stereographic transform is the
        lower (upper) pole [00-1] ([001]), i.e. only vectors with z > 0
        (z < 0) are plotted.

        Derived from :attr:`hemisphere`.
        """
        return {"upper": -1, "lower": 1}[self.hemisphere]

    @property
    def _projection(self):
        return StereographicProjection(self.pole)

    @property
    def _inverse_projection(self):
        return InverseStereographicProjection(self.pole)

    def pole_density_function(
        self,
        *args: Union[np.ndarray, Vector3d],
        resolution: float = 1,
        sigma: float = 5,
        log: bool = False,
        colorbar: bool = True,
        weights: Optional[np.ndarray] = None,
        **kwargs: Any,
    ):
        """Compute the Pole Density Function (PDF) of vectors in the
        stereographic projection. See :cite:`rohrer2004distribution`.

        Parameters
        ----------
        *args
            Vector(s), or azimuth and polar angles of the vectors, the
            latter passed as two separate arguments.
        resolution
            The angular resolution of the sampling grid in degrees.
            Default value is 1.
        sigma
            The angular resolution of the applied broadening in degrees.
            Default value is 5.
        log
            If ``True`` the log(PDF) is calculated. Default is ``True``.
        colorbar
            If ``True`` a colorbar is shown alongside the PDF plot.
            Default is ``True``.
        weights
            The weights for the individual vectors. If not given, the
            weight of each vector is 1.
        **kwargs
            Keyword arguments passed to
            :meth:`matplotlib.axes.Axes.pcolormesh`.

        See Also
        --------
        orix.measure.pole_density_function
        orix.plot.InversePoleFigurePlot.pole_density_function
        orix.vector.Vector3d.pole_density_function
        """
        hist, (x, y) = pole_density_function(
            *args,
            resolution=resolution,
            sigma=sigma,
            log=log,
            hemisphere=self.hemisphere,
            weights=weights,
        )

        new_kwargs = dict(zorder=ZORDER["mesh"], clip_on=True)
        updated_kwargs = {**kwargs, **new_kwargs}

        # plot mesh
        updated_kwargs.setdefault("cmap", "magma")
        # mpl.QuadMesh handles masked values by default
        pc = self.pcolormesh(x, y, hist, **updated_kwargs)

        if colorbar:
            label = "MRD"
            if log:
                label = f"log({label})"
            plt.colorbar(pc, label=label, ax=self)

    def draw_circle(
        self,
        *args: Union[Vector3d, Tuple[float, float], Tuple[np.ndarray, np.ndarray]],
        opening_angle: Union[float, np.ndarray] = np.pi / 2,
        steps: int = 100,
        reproject: bool = False,
        reproject_plot_kwargs: Optional[dict] = None,
        **kwargs,
    ):
        r"""Draw great or small circles with a given `opening_angle` to
        one or multiple vectors.

        A vector must be present in the current hemisphere for its
        circle to be drawn.

        Parameters
        ----------
        args
            Vector(s), or azimuth and polar angles defining vectors, the
            latter two passed as separate arguments (not keyword
            arguments). Circles are drawn perpendicular to these with a
            given ``opening_angle``.
        opening_angle
            Opening angle(s) around the vector(s). Default is
            :math:`\pi/2`, giving a great circle. If an array is passed,
            its size must be equal to the number of circles to draw.
        steps
            Number of vectors to describe each circle, default is 100.
        reproject
            Whether to reproject parts of the circle(s) visible on the
            other hemisphere. Re-projection is achieved by reflection of
            the circle(s) parts located on the other hemisphere in the
            projection plane. Ignored if ``hemisphere`` is ``"both"``.
            Default is ``False``.
        reproject_plot_kwargs
            Keyword arguments passed to
            :meth:`matplotlib.axes.Axes.plot` to alter the appearance of
            parts of the circle(s) visible on the other hemisphere if
            ``reproject=True``. These lines are dashed by default.
            Values used for circle(s) on the current hemisphere are used
            unless values are passed here.
        **kwargs
            Keyword arguments passed to
            :meth:`matplotlib.axes.Axes.plot` to alter the circles'
            appearance.

        See Also
        --------
        orix.vector.Vector3d.get_circle
        """
        out = self._prepare_to_call_inherited_method(args, kwargs)
        x, y, visible, updated_kwargs = out
        if x.size == 0:
            return

        if isinstance(opening_angle, np.ndarray) and opening_angle.size == visible.size:
            opening_angle = opening_angle[visible]

        # Get set of `steps` vectors delineating a circle per vector
        v = self._inverse_projection.xy2vector(x, y)
        circles = v.get_circle(opening_angle=opening_angle, steps=steps).unit

        # Enable using one color per circle
        color = kwargs.pop("color", "C0")
        color2 = _get_array_of_values(value=color, visible=visible)

        # Set above which plotting elements circles will appear (zorder)
        new_kwargs = dict(zorder=ZORDER["draw_circle"], clip_on=True)
        for k, v in new_kwargs.items():
            kwargs.setdefault(k, v)

        for i, c in enumerate(circles):
            self.plot(c.azimuth, c.polar, color=color2[i], **kwargs)

        if reproject:
            if reproject_plot_kwargs is None:
                reproject_plot_kwargs = {}
            reproject_plot_kwargs.setdefault("linestyle", "--")
            # Simulate reflection about the projection plane by
            # temporarily setting the hemisphere to the other one
            other_hemisphere = {"upper": "lower", "lower": "upper"}
            self._hemisphere = other_hemisphere[self._hemisphere]
            for i, c in enumerate(circles):
                self.plot(c.azimuth, c.polar, color=color2[i], **reproject_plot_kwargs)
            self._hemisphere = other_hemisphere[self._hemisphere]

    def restrict_to_sector(
        self,
        sector: FundamentalSector,
        pad: float = 1.3,
        show_edges: bool = True,
        **kwargs,
    ):
        """Restrict the stereographic axis to a
        :class:`~orix.vector.FundamentalSector`, typically obtained from
        :attr:`~orix.quaternion.Symmetry.fundamental_sector`.

        Parameters
        ----------
        sector
            Fundamental sector with edges delineating a fundamental
            sector.
        pad
            Padding outside the sector in the stereographic projection
            in degrees. Default is 1.3 degrees.
        show_edges
            Whether to draw the sector edges. Default is ``True``.
        **kwargs
            Keyword arguments passed to
            :class:`matplotlib.patches.PathPatch` if
            ``show_edges=True``.
        """
        original_pole = deepcopy(sector._pole)
        sector._pole = self.pole
        edges = sector.edges
        if edges.size == 0:
            return
        edges = _closed_edges_in_hemisphere(edges, sector, pole=self.pole)
        sector._pole = original_pole
        if edges.size == 0:
            return
        x, y, _ = self._pretransform_input((edges,))

        pad_angle = np.deg2rad(pad)
        verts = sector.vertices
        center = sector.center
        verts_normal = center.cross(verts)
        verts_rot = verts.rotate(verts_normal, pad_angle)
        x_pad, y_pad = self._projection.vector2xy(verts_rot)
        pad_min = 0.01
        if x_pad.size == 0:
            x_min, x_max = np.min(x) - pad_min, np.max(x) + pad_min
            y_min, y_max = np.min(y) - pad_min, np.max(y) + pad_min
        else:
            x_min = min([np.min(x) - pad_min, np.min(x_pad)])
            x_max = max([np.max(x) + pad_min, np.max(x_pad)])
            y_min = min([np.min(y) - pad_min, np.min(y_pad)])
            y_max = max([np.max(y) + pad_min, np.max(y_pad)])
        self.set(xlim=(x_min, x_max), ylim=(y_min, y_max))

        self.patches[0].set_visible(False)

        if show_edges:
            for k, v in [("facecolor", "none"), ("edgecolor", "k"), ("linewidth", 1)]:
                kwargs.setdefault(k, v)
            patch = mpatches.PathPatch(
                mpath.Path(np.column_stack([x, y]), closed=True),
                label="sa_sector",
                **kwargs,
            )
            self.add_patch(patch)
            self.set_clip_path(patch)
            labels = ["sa_azimuth_grid", "sa_polar_grid"]
            for c in self.collections:
                if c.get_label() in labels:
                    c.set_clip_path(patch)

    def show_hemisphere_label(self, **kwargs):
        """Add a hemisphere label (``"upper"``/``"lower"``) to the upper
        left of the plot.

        Parameters
        ----------
        **kwargs
            Keyword arguments passed to
            :func:`matplotlib.axes.Axes.text`.

        See Also
        --------
        hemisphere
        """
        new_kwargs = dict(ha="right", va="bottom")
        new_kwargs.update(kwargs)
        super().text(-0.71, 0.71, s=self.hemisphere, **new_kwargs)

    def set_labels(
        self,
        xlabel: Union[str, bool, None] = "x",
        ylabel: Union[str, bool, None] = "y",
        zlabel: Union[str, bool, None] = "z",
        **kwargs,
    ):
        """Set the reference frame's axes labels.

        Parameters
        ----------
        xlabel
            X axis label, default is ``"x"``. If ``False`` or ``None``,
            this label is not shown.
        ylabel
            Y axis label, default is ``"y"``. If ``False`` or ``None``,
            this label is not shown.
        zlabel
            Z axis label, default is ``"z"``. If ``False`` or ``None``,
            this label is not shown.
        """
        pos = [(1, 0), (0, 1), (0, 0)]
        for (x, y), label in zip(pos, [xlabel, ylabel, zlabel]):
            if label not in [None, False]:
                self._set_label(x=x, y=y, label=label, **kwargs)

    def stereographic_grid(
        self,
        show_grid: Optional[bool] = None,
        azimuth_resolution: Optional[float] = None,
        polar_resolution: Optional[float] = None,
    ):
        """Turn a stereographic grid on or off, and set the azimuth and
        polar grid resolution in degrees.

        Parameters
        ----------
        show_grid
            Whether to show grid lines. If any keyword arguments are
            passed, this is set to ``True``. If not given and there are
            no keyword arguments, the grid lines are toggled.
        azimuth_resolution
            Azimuth grid resolution in degrees. Default is 10 degrees.
            This can also be set upon initialization of the axes by
            passing ``azimuth_resolution`` to ``subplot_kw``.
        polar_resolution
            Polar grid resolution in degrees. Default is 10 degrees.
            This can also be set upon initialization of the axes by
            passing ``polar_resolution`` to ``subplot_kw``.

        See Also
        --------
        matplotlib.axes.Axes.grid
        """
        if (
            show_grid is None
            and self._stereographic_grid in [None, False]
            or show_grid is None
            and (azimuth_resolution is not None or polar_resolution is not None)
            or show_grid is True
        ) and hasattr(self, "patch"):
            self._azimuth_grid(azimuth_resolution)
            self._polar_grid(polar_resolution)
            self._stereographic_grid = True
        elif show_grid in [None, False] and self._stereographic_grid is True:
            # Remove grid
            has_azimuth, index_azimuth = self._has_collection(
                "sa_azimuth_grid", self.collections
            )
            has_polar, index_polar = self._has_collection(
                "sa_polar_grid", self.collections
            )
            if has_azimuth:
                if index_polar > index_azimuth:
                    index_polar -= 1
                self.collections[index_azimuth].remove()
            if has_polar:
                self.collections[index_polar].remove()
            self._stereographic_grid = False

    def symmetry_marker(self, v: Vector3d, fold: int, **kwargs):
        """Plot 2-, 3- 4- or 6-fold symmetry marker(s).

        Parameters
        ----------
        v
            Position of the marker(s) to plot.
        fold
            Which symmetry element to plot, can be either 2, 3, 4 or 6.
        **kwargs
            Keyword arguments passed to :meth:`scatter`.
        """
        if fold not in [2, 3, 4, 6]:
            raise ValueError("Can only plot 2-, 3-, 4- or 6-fold elements.")

        marker_classes = {
            "2": TwoFoldMarker,
            "3": ThreeFoldMarker,
            "4": FourFoldMarker,
            "6": SixFoldMarker,
        }
        marker = marker_classes[str(fold)](v, size=kwargs.pop("s", 1))

        new_kwargs = dict(zorder=ZORDER["symmetry_marker"], clip_on=False)
        for k, v in new_kwargs.items():
            kwargs.setdefault(k, v)

        for vec, marker, marker_size in marker:
            self.scatter(vec, marker=marker, s=marker_size, **kwargs)

        # TODO: Find a way to control padding, so that markers aren't
        #  clipped

    def _azimuth_grid(self, resolution: Optional[float] = None):
        """Set the azimuth grid resolution in degrees.

        Parameters
        ----------
        resolution
            Azimuth grid resolution in degrees. Default is 10 degrees.
            This can also be set upon initialization of the axes by
            passing ``azimuth_resolution`` to ``subplot_kw``.

        See Also
        --------
        polar_grid
        matplotlib.axes.Axes.grid
        """
        if resolution is not None:
            self._azimuth_resolution = resolution

        azimuth_start = np.arange(0, np.pi, np.radians(self._azimuth_resolution))
        polar = np.full(azimuth_start.size, np.pi / 2)
        if self.hemisphere == "lower":
            polar += 1e-9
        v_start = Vector3d.from_polar(azimuth_start, polar)
        x_start, y_start = self._projection.vector2xy(v_start)
        v_end = Vector3d.from_polar(azimuth_start + np.pi, polar)
        x_end, y_end = self._projection.vector2xy(v_end)

        kwargs = dict(
            linewidths=rcParams["grid.linewidth"],
            linestyle=rcParams["grid.linestyle"],
            alpha=rcParams["grid.alpha"],
            color=rcParams["grid.color"],
            antialiased=True,
            zorder=ZORDER["grid"],
        )

        label = "sa_azimuth_grid"
        lines = np.stack(((x_start, x_end), (y_start, y_end))).T
        lines_collection = mcollections.LineCollection(lines, label=label, **kwargs)
        has_collection, index = self._has_collection(label, self.collections)
        if has_collection:
            self.collections[index].remove()
        has_sector, sector_index = self._has_collection("sa_sector", self.patches)
        if has_sector:
            lines_collection.set_clip_path(self.patches[sector_index])
        self.add_collection(lines_collection)

    @staticmethod
    def _has_collection(label, collections):
        labels = [c.get_label() for c in collections]
        for i in range(len(labels)):
            if label == labels[i]:
                return True, i
        return False, -1

    def _polar_grid(self, resolution: Optional[float] = None):
        """Set the polar grid resolution in degrees.

        Parameters
        ----------
        resolution
            Polar grid resolution in degrees. Default is 15 degrees.
            This can also be set upon initialization of the axes by
            passing ``polar_resolution`` to ``subplot_kw``.

        See Also
        --------
        azimuth_grid
        matplotlib.axes.Axes.grid
        """
        if resolution is not None:
            self._polar_resolution = resolution

        res = np.radians(self._polar_resolution)

        polar = np.arange(res, np.pi, res)
        v = Vector3d.from_polar(np.zeros(polar.size), polar)
        radii, _ = self._projection.vector2xy(v)

        ec = rcParams["grid.color"]
        kwargs = dict(
            xy=(0, 0),
            linewidth=rcParams["grid.linewidth"],
            linestyle=rcParams["grid.linestyle"],
            alpha=rcParams["grid.alpha"],
            ec=ec,
            fc="none",
            antialiased=True,
            zorder=ZORDER["grid"],
        )

        circles = []
        for r in radii:
            circles.append(mpatches.Circle(radius=r, **kwargs))
        label = "sa_polar_grid"
        circles_collection = mcollections.PatchCollection(
            circles,
            label=label,
            edgecolors=kwargs["ec"],
            facecolors=kwargs["fc"],
            alpha=kwargs["alpha"],
        )
        has_collection, index = self._has_collection(label, self.collections)
        if has_collection:
            self.collections[index].remove()
        has_sector, sector_index = self._has_collection("sa_sector", self.patches)
        if has_sector:
            circles_collection.set_clip_path(self.patches[sector_index])
        self.add_collection(circles_collection)

    def _prepare_to_call_inherited_method(
        self,
        args: Union[Vector3d, Tuple[float, float], Tuple[np.ndarray, np.ndarray]],
        kwargs: dict,
        new_kwargs: Optional[dict] = None,
        sort: bool = False,
        offset: Tuple[float, float] = None,
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray, dict]:
        """Prepare arguments and keyword arguments passed to methods in
        :class:`StereographicPlot` inherited from
        :class:`matplotlib.axes.Axes`.

        Parameters
        ----------
        args
            Any arguments passed to the :class:`StereographicPlot`
            method.
        kwargs
            Any arguments passed to the :class:`StereographicPlot`
            method.
        new_kwargs
            Any default keyword arguments to be passed to the inherited
            method.
        sort
            Whether to sort vectors before passing them to Matplotlib.
            Default is ``False``.
        offset
            Tuple of offsets in (X, Y). No offset is applied if not
            given.

        Returns
        -------
        x
        y
        visible
        updated_kwargs
        """
        updated_kwargs = kwargs
        if new_kwargs is not None:
            for k, v in new_kwargs.items():
                updated_kwargs.setdefault(k, v)
        x, y, visible = self._pretransform_input(args, sort=sort, offset=offset)
        return x, y, visible, updated_kwargs

    def _pretransform_input(
        self,
        values: Union[Vector3d, Tuple[float, float], Tuple[np.ndarray, np.ndarray]],
        sort: bool = False,
        offset: Optional[Tuple[float, float]] = None,
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """Return arrays of (X, Y) from input data.

        Parameters
        ----------
        values
            Spherical coordinates (azimuth, polar) or vectors. If
            spherical coordinates are given, they are assumed to
            describe unit vectors. Vectors will be made into unit
            vectors if they are not already.
        sort
            Whether to sort vectors before passing them to Matplotlib.
            Default is ``False``.
        offset
            Tuple of offsets in (X, Y). No offset is applied if not
            given.

        Returns
        -------
        X
            Stereographic X coordinates of unit vectors.
        Y
            Stereographic Y coordinates of unit vectors.
        visible
            Whether these values are visible on the axes.
        """
        pole = self.pole
        if len(values) == 2:
            azimuth, polar = values[0], values[1]
            if sort:
                order = _order_in_hemisphere(polar, pole)
                azimuth = azimuth[order]
                polar = polar[order]
            x, y = self._projection.spherical2xy(azimuth=azimuth, polar=polar)
            v = self._inverse_projection.xy2vector(x, y)
        else:
            try:
                v = values[0].flatten().unit
                if sort:
                    order = _order_in_hemisphere(v.polar, pole)
                    v = v[order]
                x, y = self._projection.vector2xy(v)
            except (ValueError, AttributeError):
                raise ValueError(
                    "Accepts only one (Vector3d) or two (azimuth, polar) input "
                    "arguments."
                )

        visible = v <= self._projection.region

        if offset is not None:
            x += offset[0]
            y += offset[1]

        return x, y, visible

    def _set_label(self, x: float, y: float, label: str, **kwargs):
        bbox_dict = dict(boxstyle="round, pad=0.1", fc="w", ec="w")
        new_kwargs = dict(ha="center", va="center", bbox=bbox_dict)
        new_kwargs.update(kwargs)
        super().text(x=x, y=y, s=label, **new_kwargs)


mprojections.register_projection(StereographicPlot)


def _get_array_of_values(
    value: Union[str, float, List[str], List[float]], visible: np.ndarray
) -> np.ndarray:
    """Return a usable array of ``value`` with the correct size
    even though ``value`` doesn't have as many elements as
    ``visible.size``, to be iterated over along with ``True`` elements
    in ``visible``.

    Parameters
    ----------
    value
        Typically a keyword argument value to be passed to some
        Matplotlib routine.
    visible
        Boolean array with as many elements as input vectors, only some
        of which are visible in the hemisphere (``True``).

    Returns
    -------
    array
        An array populated with ``value`` of a size equal to the number
        of ``True`` elements in ``visible``.
    """
    n = visible.size
    if not isinstance(value, str) and hasattr(value, "__iter__") and len(value) != n:
        value = value[0]
    if isinstance(value, str) or not hasattr(value, "__iter__"):
        value = [value] * n
    return np.asarray(value)[visible]


def _is_visible(polar: np.ndarray, pole: int) -> np.ndarray:
    """Return a boolean array describing whether the vector which the
    polar angles belong to are visible in the current hemisphere.

    Parameters
    ----------
    polar
    pole

    Returns
    -------
    polar_visible
        Boolean array with ``True`` for polar angles corresponding to
        vectors visible in this hemisphere.
    """
    if pole == -1:
        return polar <= np.pi / 2
    else:  # pole == 1
        return polar >= np.pi / 2


def _order_in_hemisphere(polar: np.ndarray, pole: int) -> Union[np.ndarray, None]:
    """Return order of vectors based on polar angles, so that the ones
    corresponding to vectors visible in this hemisphere are shifted to
    the start of the arrays.

    Used in :meth:`StereographicPlot._pretransform_input` when
    ``sort=True``.

    Parameters
    ----------
    polar
    pole

    Returns
    -------
    out
        If no vectors are visible, ``None`` is returned.
    """
    visible = _is_visible(polar, pole)
    if visible.size == 0 or not np.any(visible):
        return

    indices = np.asarray(visible != visible[0]).nonzero()[0]
    order = np.arange(visible.size)
    if indices.size != 0:
        order = np.roll(order, shift=-(indices[-1] + 1))

    return order



r"""
===================
Combining rotations
===================

This example demonstrates how to combine two rotations :math:`R_1` and :math:`R_2`, i.e.
from right to left like so

.. math::

    R_{12} = R_2 \cdot R_1.

To convince ourselves that this order is correct, we rotate a vector
:math:`\mathbf{v} = (0, 0, z)` by two rotations: rotation :math:`R_1` by
:math:`120^{\circ}` around :math:`[1 1 1]` followed by rotation :math:`R_2` by
:math:`180^{\circ}` around :math:`[1 1 0]`.
Rotation :math:`R_1` is carried out first, followed by rotation :math:`R_2`.

Note that a negative angle :math:`\omega` when *defining* a rotation in the axis-angle
representation :math:`(\hat{\mathbf{n}}, \omega)` is necessary for consistent
transformations between rotation representations.
"""

import matplotlib.pyplot as plt

from orix import plot
from orix.quaternion import Rotation
from orix.vector import Vector3d

plt.rcParams.update({"font.size": 12, "grid.alpha": 0.5})

R1 = Rotation.from_axes_angles([1, 1, 1], -120, degrees=True)
R2 = Rotation.from_axes_angles([1, 1, 0], -180, degrees=True)
R12 = R2 * R1

R_all = Rotation.stack([R1, R2, R12]).squeeze()
print("R1, R2 and R12:\n* As quaternions:\n", R_all)
print("* As orientation matrices:\n", R_all.to_matrix().squeeze().round(10))

v_start = Vector3d.zvector()
v_end = R12 * v_start
print(
    "Point rotated by R12:\n",
    v_start.data.squeeze().tolist(),
    "->",
    v_end.data.squeeze().round(10).tolist(),
)

# Illustrate the steps of the rotation by plotting the vector before
# (red), during (green), and after (blue) the rotation and the rotation
# paths (first: cyan; second: magenta)
v_intermediate = R1 * v_start

v_si_path = Vector3d.from_path_ends(Vector3d.stack([v_start, v_intermediate]))
v_sie_path = Vector3d.from_path_ends(Vector3d.stack([v_intermediate, v_end]))

fig = plt.figure(layout="tight")
ax0 = fig.add_subplot(121, projection="stereographic")
ax1 = fig.add_subplot(122, projection="stereographic", hemisphere="lower")
ax0.stereographic_grid(), ax1.stereographic_grid()
Vector3d.stack([v_start, v_intermediate, v_end]).scatter(
    figure=fig,
    s=50,
    c=["r", "g", "b"],
    axes_labels=["e1", "e2"],
)
ax0.plot(v_si_path, color="c"), ax1.plot(v_si_path, color="c")
ax0.plot(v_sie_path, color="m"), ax1.plot(v_sie_path, color="m")
R1.axis.scatter(figure=fig, c="orange")
R2.axis.scatter(figure=fig, c="k")
text_kw = {
    "bbox": {"alpha": 0.5, "fc": "w", "boxstyle": "round,pad=0.1"},
    "ha": "right",
    "va": "top",
}
ax0.text(v_start, s="Start", **text_kw)
ax1.text(v_intermediate, s="Intermediate", **text_kw)
ax1.text(v_end, s="End", **text_kw)
ax1.text(R1.axis, s=r"$\hat{\mathbf{n}}_1$", **text_kw)
ax0.text(R2.axis, s=r"$\hat{\mathbf{n}}_2$", **text_kw)



r"""
========================
Plot symmetry operations
========================

This example shows how to draw proper symmetry operations :math:`s`
(no reflections or inversions).
"""

import matplotlib.pyplot as plt

from orix import plot
from orix.vector import Vector3d

marker_size = 200
fig, (ax0, ax1) = plt.subplots(
    ncols=2,
    subplot_kw={"projection": "stereographic"},
    layout="tight",
)

ax0.set_title("432", pad=20)
# 4-fold (outer markers will be clipped a bit...)
v4fold = Vector3d([[0, 0, 1], [1, 0, 0], [-1, 0, 0], [0, 1, 0], [0, -1, 0]])
ax0.symmetry_marker(v4fold, fold=4, c="C4", s=marker_size)
ax0.draw_circle(v4fold, color="C4")
# 3-fold
v3fold = Vector3d([[1, 1, 1], [1, -1, 1], [-1, -1, 1], [-1, 1, 1]])
ax0.symmetry_marker(v3fold, fold=3, c="C3", s=marker_size)
ax0.draw_circle(v3fold, color="C3")
# 2-fold
# fmt: off
v2fold = Vector3d(
    [
        [ 1,  0, 1],
        [ 0,  1, 1],
        [-1,  0, 1],
        [ 0, -1, 1],
        [ 1,  1, 0],
        [-1, -1, 0],
        [-1,  1, 0],
        [ 1, -1, 0],
    ]
)
# fmt: on
ax0.symmetry_marker(v2fold, fold=2, c="C2", s=marker_size)
ax0.draw_circle(v2fold, color="C2")

ax1.set_title("222", pad=20)
# 2-fold
v2fold = Vector3d([[0, 0, 1], [1, 0, 0], [-1, 0, 0], [0, 1, 0], [0, -1, 0]])
ax1.symmetry_marker(v2fold, fold=2, c="C2", s=2 * marker_size)
ax1.draw_circle(v2fold, color="C2")



# -*- coding: utf-8 -*-
# Copyright 2018-2024 the orix developers
#
# This file is part of orix.
#
# orix is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# orix is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with orix.  If not, see <http://www.gnu.org/licenses/>.

"""Uniform cubochoric sampling of *SO(3)* :cite:`singh2016orientation`.

This module and documentation is only relevant for orix developers, not
for users.

.. warning:
    This module is for internal use only.  Do not use it in your own
    code. We may change the API at any time with no warning.
"""

from typing import Optional, Union

import numba as nb
import numpy as np

from orix.quaternion import Rotation
from orix.quaternion._conversions import ax2qu_single, cu2ro_single, ro2ax_single


def cubochoric_sampling(
    semi_edge_steps: Optional[int] = None,
    resolution: Optional[Union[int, float]] = None,
) -> Rotation:
    r"""Uniform cubochoric sampling of rotations *SO(3)*
    :cite:`singh2016orientation`.

    Parameters
    ----------
    semi_edge_steps
        Number of grid points :math:`N` along the semi-edge of the
        cubochoric cube. If not given, it will be calculated from
        ``resolution`` following Eq. (9) in
        :cite:`singh2016orientation`. For example, if an average
        disorientation of 1 degree is needed, this should be set to 137.
        This will result in :math:`(2N + 1)^3 = 20 796 875` unique
        rotations.
    resolution
        Average disorientation between resulting rotations. This must be
        given if ``semi_edge_steps`` is not.

    Returns
    -------
    rot
        Sampled rotations in *SO(3)*.

    Notes
    -----
    The cubochoric grid sampled is :math:`S_{000}(N)`, which contains
    the identity rotation.
    """
    if semi_edge_steps is None:
        if resolution is None:
            raise ValueError("Either `semi_edge_steps` or `resolution` must be passed")
        else:
            semi_edge_steps = resolution_to_semi_edge_steps(resolution)
    quaternions = _cubochoric_sampling_loop(semi_edge_steps)
    return Rotation(quaternions)


@nb.jit(cache=True, nogil=True, nopython=True)
def resolution_to_semi_edge_steps(resolution: Union[int, float]) -> int:
    r"""Calculate the number of grid points :math:`N` along the
    semi-edge of the cubochoric cube given an average disorientation
    between rotations :cite:`singh2016orientation`.

    Parameters
    ----------
    resolution
        Resolution in degrees.

    Returns
    -------
    steps
        Semi-edge steps.
    """
    return int(np.round(131.97049 / (resolution - 0.03732)))


@nb.jit("float64[:, :](int64)", cache=True, nogil=True, nopython=True)
def _cubochoric_sampling_loop(semi_edge_steps: int) -> np.ndarray:
    """See :func:`cubochoric_sampling`.

    If ``semi_edge_steps`` is 100, there will be (201, 201, 201) points
    sampled.
    """
    semi_edge_length = 0.5 * np.pi ** (2 / 3)
    step_size = semi_edge_length / semi_edge_steps
    n_points = (2 * semi_edge_steps + 1) ** 3
    rot = np.zeros((n_points, 4))

    xyz = np.zeros(3)
    step = 0
    for i in nb.prange(-semi_edge_steps + 1, semi_edge_steps + 1):
        xyz[0] = i * step_size
        for j in range(-semi_edge_steps + 1, semi_edge_steps + 1):
            xyz[1] = j * step_size
            for k in range(-semi_edge_steps + 1, semi_edge_steps + 1):
                xyz[2] = k * step_size

                # Discard the point and move to the next iteration if it
                # lies outside the cubochoric cube
                if np.max(np.abs(xyz)) > semi_edge_length:  # pragma: no cover
                    continue

                # Get quaternion via cubochoric coordinates -> Rodrigues
                # vector -> axis-angle pair
                rodrigues = cu2ro_single(xyz)
                axis_angle = ro2ax_single(rodrigues)
                rot[step] = ax2qu_single(axis_angle)

                step += 1

    return rot[:step]



"""
=========
Wulff net
=========

This example shows how to draw a Wulff net in the stereographic projection with great
and small circles.
"""

import matplotlib.pyplot as plt
import numpy as np

from orix import plot
from orix.vector import Vector3d

n = int(90 / 2)  # Degree / net resolution
steps = 500
kwargs = dict(linewidth=0.25, color="k")

polar = np.linspace(0, 0.5 * np.pi, num=n)
v_right = Vector3d.from_polar(azimuth=np.zeros(n), polar=polar)
v_left = Vector3d.from_polar(azimuth=np.ones(n) * np.pi, polar=polar)
v010 = Vector3d.zero(shape=(n,))
v010.y = 1
v010_opposite = -v010

fig, ax = plt.subplots(
    figsize=(5, 5), subplot_kw=dict(projection="stereographic"), layout="tight"
)
ax.stereographic_grid(False)
ax.draw_circle(v_right, steps=steps, **kwargs)
ax.draw_circle(v_left, steps=steps, **kwargs)
ax.draw_circle(v010, opening_angle=polar, steps=steps, **kwargs)
ax.draw_circle(v010_opposite, opening_angle=polar, steps=steps, **kwargs)
for label, azimuth, va, ha, offset in zip(
    ["B", "M''", "A", "M'"],
    np.array([0, 0.5, 1, 1.5]) * np.pi,
    ["center", "bottom", "center", "top"],
    ["left", "center", "right", "center"],
    [(0.02, 0), (0, 0.02), (-0.02, 0), (0, -0.02)],
):
    ax.text(azimuth, 0.5 * np.pi, s=label, offset=offset, c="r", va=va, ha=ha)



# -*- coding: utf-8 -*-
# Copyright 2018-2024 the orix developers
#
# This file is part of orix.
#
# orix is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# orix is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with orix.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np
import pytest

from orix.quaternion import Orientation
from orix.quaternion._conversions import (
    ax2qu,
    ax2qu_2d,
    ax2qu_single,
    ax2ro,
    ax2ro_2d,
    ax2ro_single,
    cu2ho,
    cu2ho_2d,
    cu2ho_single,
    cu2ro,
    cu2ro_2d,
    cu2ro_single,
    eu2qu,
    eu2qu_2d,
    eu2qu_single,
    get_pyramid,
    get_pyramid_2d,
    get_pyramid_single,
    ho2ax,
    ho2ax_2d,
    ho2ax_single,
    ho2ro,
    ho2ro_2d,
    ho2ro_single,
    om2qu,
    om2qu_3d,
    om2qu_single,
    qu2ax,
    qu2ax_2d,
    qu2ax_single,
    qu2eu,
    qu2eu_2d,
    qu2eu_single,
    qu2ho,
    qu2ho_2d,
    qu2ho_single,
    qu2om,
    qu2om_2d,
    qu2om_single,
    ro2ax,
    ro2ax_2d,
    ro2ax_single,
)
from orix.quaternion.symmetry import Oh


class TestRotationConversions:
    """Tests of conversions between rotation representations.

    Functions are accelerated with Numba, so both the Numba and Python
    versions of the functions are tested.
    """

    def test_eu2qu2eu(self, euler_angles, quaternions_conversions):
        qu_64 = quaternions_conversions
        eu_64 = euler_angles
        qu_32_nd = qu_64.astype(np.float32).reshape(2, 5, 4)
        eu_32_nd = eu_64.astype(np.float32).reshape(2, 5, 3)

        # Single
        for qu, eu in zip(qu_64, eu_64):
            assert np.allclose(qu2eu_single.py_func(qu), eu, atol=1e-4)
            assert np.allclose(eu2qu_single.py_func(eu), qu, atol=1e-4)

        # 2D
        assert np.allclose(qu2eu_2d.py_func(qu_64), eu_64, atol=1e-4)
        assert np.allclose(eu2qu_2d.py_func(eu_64), qu_64, atol=1e-4)

        # ND and jit
        qu_64_nd = qu_64.reshape(2, 5, 4)
        eu_64_nd = eu_64.reshape(2, 5, 3)
        assert np.allclose(eu2qu(eu_64_nd), qu_64_nd, atol=1e-4)
        assert np.allclose(qu2eu(qu_64_nd), eu_64_nd, atol=1e-4)

        # ND float32
        assert np.allclose(qu2eu(qu_32_nd), eu_32_nd, atol=1e-4)
        assert np.allclose(eu2qu(eu_32_nd), qu_32_nd, atol=1e-4)

        # Symmetry-preserving
        ori = Orientation.from_euler(eu_64[-1], Oh)
        assert ori.symmetry == Oh
        assert np.allclose(ori.data[0], qu_64[-1], atol=1e-4)

    def test_get_pyramid(self, cubochoric_coordinates):
        """Cubochoric coordinates situated in expected pyramid."""
        cu_64 = cubochoric_coordinates
        pyramid = np.array([3, 1, 3, 1, 1, 2, 3, 4, 5, 6])

        # Single
        for xyz, p in zip(cu_64, pyramid):
            assert get_pyramid_single.py_func(xyz) == p

        # 2D
        assert np.allclose(get_pyramid_2d.py_func(cu_64), pyramid)

        # ND
        cu_64_nd = cu_64.reshape(5, 2, 3)
        assert np.allclose(get_pyramid(cu_64_nd), pyramid)

    def test_cu2ho(self, cubochoric_coordinates, homochoric_vectors):
        cu_64 = cubochoric_coordinates
        ho_64 = homochoric_vectors

        # Single
        for cu, ho in zip(cu_64, ho_64):
            assert np.allclose(cu2ho_single.py_func(cu), ho, atol=1e-4)

        # 2D
        assert np.allclose(cu2ho_2d.py_func(cu_64), ho_64, atol=1e-4)

        # ND
        cu_64_nd = cu_64.reshape(5, 2, 3)
        ho_64_nd = ho_64.reshape(5, 2, 3)
        assert np.allclose(cu2ho(cu_64_nd), ho_64_nd, atol=1e-4)

        # ND float32
        cu_32_nd = cu_64_nd.astype(np.float32)
        ho_32_nd = ho_64_nd.astype(np.float32)
        assert np.allclose(cu2ho(cu_32_nd), ho_32_nd, atol=1e-4)

    def test_ho2ax(self, homochoric_vectors, axis_angle_pairs):
        ho_64 = homochoric_vectors
        ax_64 = axis_angle_pairs

        # Single
        for ho, ax in zip(ho_64, ax_64):
            assert np.allclose(ho2ax_single.py_func(ho), ax, atol=1e-4)

        # 2D
        assert np.allclose(ho2ax_2d.py_func(ho_64), ax_64, atol=1e-4)

        # ND
        ho_64_nd = ho_64.reshape(2, 5, 3)
        ax_64_nd = ax_64.reshape(2, 5, 4)
        assert np.allclose(ho2ax(ho_64_nd), ax_64_nd, atol=1e-4)

        # ND float32
        ho_32_nd = ho_64_nd.astype(np.float32)
        ax_32_nd = ax_64_nd.astype(np.float32)
        assert np.allclose(ho2ax(ho_32_nd), ax_32_nd, atol=1e-4)

    def test_ax2ro2ax(
        self, axis_angle_pairs, rodrigues_vectors, quaternions_conversions
    ):
        ax_64 = axis_angle_pairs
        ro_64 = rodrigues_vectors

        # Single
        for ax, ro in zip(ax_64, ro_64):
            assert np.allclose(ax2ro_single.py_func(ax), ro, atol=1e-4)
            assert np.allclose(ro2ax_single.py_func(ro), ax, atol=1e-4)

        # 2D
        assert np.isinf(ax2ro_single.py_func(np.array([0, 0, 0, np.pi]))[3])
        assert np.allclose(ax2ro_2d.py_func(ax_64), ro_64, atol=1e-4)
        assert np.allclose(ro2ax_2d.py_func(ro_64), ax_64, atol=1e-4)

        # ND
        ax_64_nd = ax_64.reshape(2, 5, 4)
        ro_64_nd = ro_64.reshape(2, 5, 4)
        assert np.allclose(ax2ro(ax_64_nd), ro_64_nd, atol=1e-4)
        assert np.allclose(ro2ax(ro_64_nd), ax_64_nd, atol=1e-4)

        # ND float32
        ax_32_nd = ax_64_nd.astype(np.float32)
        ro_32_nd = ro_64_nd.astype(np.float32)
        assert np.allclose(ax2ro(ax_32_nd), ro_32_nd, atol=1e-4)
        assert np.allclose(ro2ax(ro_32_nd), ax_32_nd, atol=1e-4)

    def test_ax2qu2ax(self, axis_angle_pairs, quaternions_conversions):
        ax_64 = axis_angle_pairs
        qu_64 = quaternions_conversions

        # Single
        for ax, qu in zip(ax_64, qu_64):
            assert np.allclose(ax2qu_single.py_func(ax), qu, atol=1e-4)
            assert np.allclose(qu2ax_single.py_func(qu), ax, atol=2e-4)

        # Test conversion of lower hemisphere quaternion
        qu_lower = np.array([-1, 1, 1, 0], dtype=np.float64)
        ax_lower = ax2qu_single(qu_lower)
        assert np.allclose(ax_lower, np.array([1, 0, 0, 0]), atol=1e-4)

        # 2D
        assert np.allclose(ax2qu_2d.py_func(ax_64), qu_64, atol=1e-4)
        assert np.allclose(qu2ax_2d.py_func(qu_64), ax_64, atol=2e-4)

        # ND
        ax_64_nd = ax_64.reshape(2, 5, 4)
        qu_64_nd = qu_64.reshape(2, 5, 4)
        assert np.allclose(
            ax2qu(ax_64_nd[..., :3], ax_64_nd[..., 3]), qu_64_nd, atol=1e-4
        )
        ax_64_nd_out = np.dstack(qu2ax(qu_64_nd))
        assert np.allclose(ax_64_nd_out, ax_64_nd, atol=2e-4)

        # ND float32
        ax_32_nd = ax_64_nd.astype(np.float32)
        qu_32_nd = qu_64_nd.astype(np.float32)
        axes_32_nd = ax_32_nd[..., :3]
        angles_32_nd = ax_32_nd[..., 3]
        assert np.allclose(ax2qu(axes_32_nd, angles_32_nd), qu_32_nd, atol=1e-4)
        ax_32_nd_out = np.dstack(qu2ax(qu_32_nd))
        assert np.allclose(ax_32_nd_out, ax_32_nd, atol=2e-4)

        # Make sure bad data causes the expected errors
        with pytest.raises(ValueError, match="Final dimension of axes array must be 3"):
            ax2qu(axes_32_nd.T, angles_32_nd)
        with pytest.raises(
            ValueError,
            match=r"The dimensions of axes \(2, 3\) and angles \(2, 2\) are ",
        ):
            ax2qu(axes_32_nd[:, :3], angles_32_nd[:, :2])
        with pytest.raises(
            ValueError, match="Final dimension of quaternion array must be 4"
        ):
            qu2ax(qu_64[:, :3])

    def test_qu2ax_negative_magnitude(self):
        q = np.array([-0.1, 0.2, 0.3, 0.4])
        ax = qu2ax_single.py_func(q)
        assert np.allclose(ax, [-0.3714, -0.5571, -0.7428, 3.3419], atol=1e-4)

    def test_ho2ro(self, homochoric_vectors, rodrigues_vectors):
        ho_64 = homochoric_vectors
        ro_64 = rodrigues_vectors

        # Single
        for ho, ro in zip(ho_64, ro_64):
            assert np.allclose(ho2ro_single.py_func(ho), ro, atol=1e-4)

        # 2D
        assert np.allclose(ho2ro_2d.py_func(ho_64), ro_64, atol=1e-4)

        # ND
        ho_64_nd = ho_64.reshape(2, 5, 3)
        ro_64_nd = ro_64.reshape(2, 5, 4)
        assert np.allclose(ho2ro(ho_64_nd), ro_64_nd, atol=1e-4)

        # ND float32
        ho_32_nd = ho_64_nd.astype(np.float32)
        rod_32_nd = ro_64_nd.astype(np.float32)
        assert np.allclose(ho2ro(ho_32_nd), rod_32_nd, atol=1e-4)

    def test_cu2ro(self, cubochoric_coordinates, rodrigues_vectors):
        cu_64 = cubochoric_coordinates
        ro_64 = rodrigues_vectors

        # Single
        for cu, ro in zip(cu_64, ro_64):
            assert np.allclose(cu2ro_single.py_func(cu), ro, atol=1e-4)

        # 2D
        assert np.allclose(cu2ro_2d.py_func(cu_64), ro_64, atol=1e-4)

        # ND
        cu_64_nd = cu_64.reshape(2, 5, 3)
        ro_64_nd = ro_64.reshape(2, 5, 4)
        assert np.allclose(cu2ro(cu_64_nd), ro_64_nd, atol=1e-4)

        # ND float32
        cu_32_nd = cu_64_nd.astype(np.float32)
        ro_32_nd = ro_64_nd.astype(np.float32)
        assert np.allclose(cu2ro(cu_32_nd), ro_32_nd, atol=1e-4)

    def test_om2qu2om(self, orientation_matrices, quaternions_conversions):
        # Checks both om2qu and qu2om simultaneously
        om_64 = orientation_matrices
        qu_64 = quaternions_conversions

        # Single
        for om, qu in zip(om_64, qu_64):
            assert np.allclose(om2qu_single.py_func(om), qu, atol=1e-4)
            assert np.allclose(qu2om_single.py_func(qu), om, atol=1e-4)

        # Test edge cases where q[0] is at or near zero
        rot_pi_111 = np.array([[-1, 2, 2], [2, -1, 2], [2, 2, -1]]) / 3
        qu_from_om = om2qu_single.py_func(rot_pi_111)
        qu_actual = np.array([0, np.sqrt(1 / 3), np.sqrt(1 / 3), np.sqrt(1 / 3)])
        om_from_qu = qu2om_single.py_func(qu_actual)
        assert np.allclose(qu_from_om, qu_actual)
        assert np.allclose(om_from_qu, rot_pi_111)

        # 2D
        assert np.allclose(om2qu_3d.py_func(om_64), qu_64, atol=1e-4)
        assert np.allclose(om_64, qu2om_2d.py_func(qu_64), atol=1e-4)

        # ND
        om_64_nd = om_64.reshape(2, 5, 3, 3)
        qu_64_nd = qu_64.reshape(2, 5, 4)
        assert np.allclose(om2qu(om_64_nd), qu_64_nd, atol=1e-4)
        assert np.allclose(om_64_nd, qu2om(qu_64_nd), atol=1e-4)

        # ND float32
        om_32_nd = om_64_nd.astype(np.float32)
        qu_32_nd = qu_64_nd.astype(np.float32)
        assert np.allclose(om2qu(om_32_nd), qu_32_nd, atol=1e-4)
        assert np.allclose(om_32_nd, qu2om(qu_32_nd), atol=1e-4)

    def test_om2qu(self, orientation_matrices, quaternions_conversions):
        om_64 = orientation_matrices
        qu_64 = quaternions_conversions

        # Single
        for om, qu in zip(om_64, qu_64):
            assert np.allclose(om2qu_single.py_func(om), qu, atol=1e-4)

        # Test edge cases where q[0] is at or near zero
        rot_pi_111 = np.array([[-1, 2, 2], [2, -1, 2], [2, 2, -1]]) / 3
        qu_from_om = om2qu_single.py_func(rot_pi_111)
        qu_actual = np.array([0, np.sqrt(1 / 3), np.sqrt(1 / 3), np.sqrt(1 / 3)])
        assert np.allclose(qu_from_om, qu_actual)

        # 2D
        assert np.allclose(om2qu_3d.py_func(om_64), qu_64, atol=1e-4)

        # ND
        om_64_nd = om_64.reshape(2, 5, 3, 3)
        qu_64_nd = qu_64.reshape(2, 5, 4)
        assert np.allclose(om2qu(om_64_nd), qu_64_nd, atol=1e-4)

        # ND float32
        om_32_nd = om_64_nd.astype(np.float32)
        qu_32_nd = qu_64_nd.astype(np.float32)
        assert np.allclose(om2qu(om_32_nd), qu_32_nd, atol=1e-4)

    def test_qu2ho(self, quaternions_conversions, homochoric_vectors):
        qu_64 = quaternions_conversions
        ho_64 = homochoric_vectors

        # Single
        for qu, ho in zip(qu_64, ho_64):
            assert np.allclose(qu2ho_single.py_func(qu), ho, atol=1e-4)

        # 2D
        assert np.allclose(qu2ho_2d.py_func(qu_64), ho_64, atol=1e-4)

        # ND
        qu_64_nd = qu_64.reshape(2, 5, 4)
        ho_64_nd = ho_64.reshape(2, 5, 3)
        assert np.allclose(qu2ho(qu_64_nd), ho_64_nd, atol=1e-4)

        # ND float32
        qu_32_nd = qu_64_nd.astype(np.float32)
        ho_32_nd = ho_64_nd.astype(np.float32)
        assert np.allclose(qu2ho(qu_32_nd), ho_32_nd, atol=1e-4)

        with pytest.raises(
            ValueError, match="Final dimension of quaternion array must be 4"
        ):
            qu2ho(qu_32_nd[..., :3])

    def test_qu2ho2ax2qu(self, quaternions_conversions):
        qu1 = quaternions_conversions
        ho = qu2ho(qu1)
        ax = ho2ax(ho)
        qu2 = ax2qu(ax[..., :3], ax[..., 3])
        assert np.allclose(qu1, qu2, atol=1e-3)  # 1e-4 errors



# -*- coding: utf-8 -*-
# Copyright 2018-2024 the orix developers
#
# This file is part of orix.
#
# orix is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# orix is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with orix.  If not, see <http://www.gnu.org/licenses/>.

from typing import List, Tuple

import numpy as np
from scipy.spatial import cKDTree


def _get_start_and_end_index(
    number_of_steps: int,
    include_start: bool,
    include_end: bool,
    positive_and_negative: bool,
) -> Tuple[int, int]:
    if positive_and_negative:
        start = -number_of_steps
    else:
        start = 0
    if not include_start:
        start += 1
    end = number_of_steps
    if include_end:
        end += 1
    return start, end


def _number_of_equidistant_steps(resolution: float, length: float) -> int:
    maximum_grid_spacing = np.tan(np.deg2rad(resolution))
    return int(np.ceil(length / maximum_grid_spacing))


def _sample_length_equidistant(
    number_of_steps: int,
    length: float,
    include_start: bool = True,
    include_end: bool = False,
    positive_and_negative: bool = True,
) -> np.ndarray:
    start_index, end_index = _get_start_and_end_index(
        number_of_steps,
        include_start,
        include_end,
        positive_and_negative,
    )
    grid_spacing = length / number_of_steps
    grid_on_edge = np.arange(start_index, end_index) * grid_spacing
    return grid_on_edge


def _number_of_equiangular_steps(resolution: float, length: float) -> int:
    total_angle = np.arctan(length)
    return int(np.ceil(total_angle / np.deg2rad(resolution)))


def _sample_length_equiangular(
    number_of_steps: int,
    length: float,
    include_start: bool = True,
    include_end: bool = False,
    positive_and_negative: bool = True,
) -> np.ndarray:
    total_angle = np.arctan(length)
    start_index, end_index = _get_start_and_end_index(
        number_of_steps,
        include_start,
        include_end,
        positive_and_negative,
    )
    linear_grid = np.arange(start_index, end_index)
    angular_increment = total_angle / number_of_steps
    grid_on_edge = np.tan(linear_grid * angular_increment)
    return grid_on_edge


def _edge_grid_normalized_cube(resolution: float) -> np.ndarray:
    center_of_face_to_edge = 1.0
    number_of_steps = _number_of_equidistant_steps(resolution, center_of_face_to_edge)
    return _sample_length_equidistant(number_of_steps, center_of_face_to_edge)


def _edge_grid_spherified_edge_cube(resolution: float) -> np.ndarray:
    center_of_face_to_edge = 1.0
    number_of_steps = _number_of_equiangular_steps(resolution, center_of_face_to_edge)
    return _sample_length_equiangular(number_of_steps, center_of_face_to_edge)


def _edge_grid_spherified_corner_cube(resolution: float) -> np.ndarray:
    center_of_face_to_corner = np.sqrt(2)
    number_of_steps = _number_of_equiangular_steps(resolution, center_of_face_to_corner)
    grid_on_diagonal = _sample_length_equiangular(
        number_of_steps, center_of_face_to_corner
    )
    return grid_on_diagonal / center_of_face_to_corner


def _compose_from_faces(
    corners: np.ndarray,
    faces: List[Tuple[int, int, int]],
    n: int,
) -> np.ndarray:
    """Refine a grid starting from a platonic solid; adapted from
    :mod:`meshzoo` :cite:`meshzoo`.

    Parameters
    ----------
    corners
        Coordinates of vertices for starting shape. Shape of the array
        should be (N, 3).
    faces
        Each tuple in the list corresponds to the vertex indices making
        up a triangular face of the mesh.
    n
        Number of times the mesh is refined.

    Returns
    -------
    vertices
        The coordinates of the refined mesh vertices, an array of shape
        (N, 3).
    """
    # create corner nodes
    vertices = [corners]

    # create edges
    edges = set()
    for face in faces:
        edges.add(tuple(sorted([face[0], face[1]])))
        edges.add(tuple(sorted([face[1], face[2]])))
        edges.add(tuple(sorted([face[2], face[0]])))

    # create edge nodes:
    t = np.linspace(1 / n, 1.0, n - 1, endpoint=False)
    for edge in edges:
        i0, i1 = edge
        new_vertices = np.outer(1 - t, corners[i0]) + np.outer(t, corners[i1])
        vertices.append(new_vertices)

    for face in faces:
        face_corner_indices = face

        bary = (
            np.hstack(
                [[np.full(n - i - 1, i), np.arange(1, n - i)] for i in range(1, n)]
            )
            / n
        )
        bary = np.array([1.0 - bary[0] - bary[1], bary[1], bary[0]])
        corner_verts = np.array([corners[i] for i in face_corner_indices])
        vertices_cart = np.dot(corner_verts.T, bary).T

        vertices.append(vertices_cart)

    return np.concatenate(vertices)


def _get_first_nearest_neighbors(
    points: np.ndarray,
    leaf_size: int = 50,
) -> np.ndarray:
    """Get array of first nearest neighbor points for all points in a
    point cloud.

    Parameters
    ----------
    points
        Point cloud with shape (N, D) representing N points in D
        dimensions.
    leaf_size
        The NN search is performed using a cKDTree object. The way this
        tree is constructed depends on ``leaf_size``, so this parameter
        will influence speed of tree construction and search.

    Returns
    -------
    nn1_vec
        Point cloud represented by an array of shape (N, D) with N
        points in D dimensions, representing the nearest neighbor point
        of each point in ``points``.
    """
    tree = cKDTree(points, leaf_size)
    # get the indexes of the first nearest neighbor of each vertex
    nn1 = tree.query(points, k=2)[1][:, 1]
    nn1_vec = points[nn1]
    return nn1_vec


def _get_angles_between_nn_gridpoints(
    vertices: np.ndarray,
    leaf_size: int = 50,
) -> np.ndarray:
    """Return angles between all nearest neighbor grid points on unit
    sphere.
    """
    # normalize the vertex vectors
    vertices = (vertices.T / np.linalg.norm(vertices, axis=1)).T
    nn1_vec = _get_first_nearest_neighbors(vertices, leaf_size)
    # the dot product between each point and its nearest neighbor
    nn_dot = np.sum(vertices * nn1_vec, axis=1)
    # angles
    angles = np.rad2deg(np.arccos(nn_dot))
    return angles


def _get_max_grid_angle(
    vertices: np.ndarray,
    leaf_size: int = 50,
) -> np.ndarray:
    """Get the maximum angle between nearest neighbor grid points on a
    unit sphere.
    """
    return np.max(_get_angles_between_nn_gridpoints(vertices, leaf_size))



# -*- coding: utf-8 -*-
# Copyright 2018-2024 the orix developers
#
# This file is part of orix.
#
# orix is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# orix is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with orix.  If not, see <http://www.gnu.org/licenses/>.

from matplotlib import pyplot as plt
import numpy as np
import pytest

from orix.plot import AxAnglePlot, RodriguesPlot, RotationPlot
from orix.quaternion import Misorientation, Orientation, OrientationRegion
from orix.quaternion.symmetry import C1, D6


def test_init_rodrigues_plot():
    fig = plt.figure()
    ax = fig.add_subplot(projection="rodrigues", auto_add_to_figure=False)
    assert isinstance(ax, RodriguesPlot)


def test_init_axangle_plot():
    fig = plt.figure()
    ax = fig.add_subplot(projection="axangle", auto_add_to_figure=False)
    assert isinstance(ax, AxAnglePlot)


def test_RotationPlot_methods():
    """This code is lifted from demo-3-v0.1."""
    misori = Misorientation([1, 1, 1, 1])  # any will do
    ori = Orientation.random()
    fig = plt.figure()
    ax = fig.add_subplot(
        projection="axangle", proj_type="ortho", auto_add_to_figure=False
    )
    ax.scatter(misori)
    ax.scatter(ori)
    ax.plot(misori)
    ax.plot(ori)
    ax.plot_wireframe(OrientationRegion.from_symmetry(D6, D6))
    plt.close("all")

    # Clear the edge case
    ax.transform(np.asarray([1, 1, 1]))


def test_full_region_plot():
    empty = OrientationRegion.from_symmetry(C1, C1)
    _ = empty.get_plot_data()


def test_RotationPlot_transform_fundamental_zone_raises():
    fig = plt.figure()
    ax = RotationPlot(fig)
    fig.add_axes(ax)
    with pytest.raises(
        TypeError, match="fundamental_zone is not an OrientationRegion object"
    ):
        ax.transform(Orientation.random(), fundamental_zone=1)


def test_RotationPlot_map_into_symmetry_reduced_zone():
    # orientations are (in, out) of D6 fundamental zone
    ori = Orientation(((1, 0, 0, 0), (0.5, 0.5, 0.5, 0.5)))
    ori.symmetry = D6
    fz = OrientationRegion.from_symmetry(ori.symmetry)
    assert np.allclose(ori < fz, (True, False))
    # test map_into_symmetry_reduced_zone in RotationPlot.transform
    fig = ori.scatter(return_figure=True)
    xyz_symmetry = fig.axes[0].collections[1]._offsets3d
    # compute same plot again but with C1 symmetry where both orientations are in C1 FZ
    ori.symmetry = C1
    fig2 = ori.scatter(return_figure=True)
    xyz = fig2.axes[0].collections[1]._offsets3d
    # test that the plotted points are not the same
    assert not np.allclose(xyz_symmetry, xyz)


def test_correct_aspect_ratio():
    # Set up figure the "old" way
    fig = plt.figure()
    ax = fig.add_subplot(
        projection="axangle", proj_type="ortho", auto_add_to_figure=False
    )

    # Check aspect ratio
    x_old, _, z_old = ax.get_box_aspect()
    assert np.allclose(x_old / z_old, 1.334, atol=1e-3)

    fr = OrientationRegion.from_symmetry(D6)
    ax._correct_aspect_ratio(fr, set_limits=False)

    x_new, _, z_new = ax.get_box_aspect()
    assert np.allclose(x_new / z_new, 3, atol=1e-3)

    # Check data limits
    assert np.allclose(ax.get_xlim(), [0, 1])
    ax._correct_aspect_ratio(fr)  # set_limits=True is default
    assert np.allclose(ax.get_xlim(), [-np.pi / 2, np.pi / 2])



# %%
r"""
====================================
Orientation from aligning directions
====================================

This example demonstrates how to use
:meth:`~orix.quaternion.Orientation.from_align_vectors` to estimate an orientation
:math:`O` from two sets of aligned vectors.
One set of vectors :math:`\mathbf{v}` is given in the sample reference reference frame,
:math:`(x, y, z)`, the other set :math:`\mathbf{t}` is given in the crystal reference
frame, :math:`(e_1, e_2, e_3)`.
"""

from diffpy.structure import Lattice, Structure
import matplotlib.pyplot as plt
import numpy as np

from orix.crystal_map import Phase
from orix.quaternion import Orientation
from orix.vector import Miller, Vector3d

plt.rcParams.update({"figure.figsize": (5, 5), "lines.markersize": 8})

# Specify an hexagonal crystal structure and symmetry
phase = Phase(
    point_group="6/mmm",
    structure=Structure(lattice=Lattice(1, 1, 2, 90, 90, 120)),
)

# Define a reference orientation (goal)
O_ref = Orientation.from_axes_angles([1, 1, 1], -45, phase.point_group, degrees=True)

# Specify two crystal directions (any will do)
t = Miller(uvw=[[2, 1, 1], [1, 3, 1]], phase=phase)

# Find out where these directions in the reference orientation (crystal)
# point in the sample reference frame
v = Vector3d(~O_ref * t)

# Plot the reference orientation sample directions as empty circles
fig = v.scatter(
    c="none",
    ec=["r", "b"],
    grid=True,
    axes_labels=["X", "Y"],
    return_figure=True,
    figure_kwargs={"layout": "tight", "figsize": (5, 5)},
)

# Add some randomness to the sample directions (0 error magnitude gives
# exact result)
err_magnitude = 0.1
v_err = v + Vector3d(np.random.normal(0, err_magnitude, 3))
angle_err = v_err.angle_with(v, degrees=True).mean()
print("Vector angle deviation [deg]: ", angle_err)

# Obtain the orientation which aligns the crystal directions with the
# sample directions
O_new, err = Orientation.from_align_vectors(t, v_err, return_rmsd=True)
print("Error distance: ", err)

# Plot the crystal directions in the new orientation
v2 = Vector3d(~O_new * t)
v2.scatter(c=["r", "b"], figure=fig)



# -*- coding: utf-8 -*-
# Copyright 2018-2024 the orix developers
#
# This file is part of orix.
#
# orix is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# orix is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with orix.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np
import pytest

from orix import plot
from orix.vector import Vector3d


class TestFormatVectorLabels:
    @pytest.mark.parametrize(
        "kwargs, desired",
        [
            ({}, ["$111$", "$\\bar{2}01$", "$400$", "$\\bar{4}00$"]),
            (
                dict(brackets=("[", "]"), use_latex=False),
                ["[111]", "[-201]", "[400]", "[-400]"],
            ),
            (
                dict(brackets=("{", "}")),
                [
                    "$\\{111\\}$",
                    "$\\{\\bar{2}01\\}$",
                    "$\\{400\\}$",
                    "$\\{\\bar{4}00\\}$",
                ],
            ),
            (
                dict(brackets=("{", "}"), use_latex=False),
                ["{111}", "{-201}", "{400}", "{-400}"],
            ),
            (
                dict(brackets=("<", ">")),
                [
                    "$\\left<111\\right>$",
                    "$\\left<\\bar{2}01\\right>$",
                    "$\\left<400\\right>$",
                    "$\\left<\\bar{4}00\\right>$",
                ],
            ),
        ],
    )
    def test_format_vector_labels(self, kwargs, desired):
        v = Vector3d([[1, 1, 1], [-2, 0, 1], [4, 0, 0], [-4, 0, 0]])
        v = v.reshape(2, 2)
        labels = plot.format_labels(v.data, **kwargs)
        assert labels.shape == (2, 2)
        assert labels.flatten().tolist() == desired

    def test_format_vector_labels_4(self):
        assert all(
            plot.format_labels([[1, 2, 3, 4], [1.1, 2.2, 3.3, 4.51]])
            == np.array(["$1234$", "$1235$"], dtype="U6")
        )



# -*- coding: utf-8 -*-
# Copyright 2018-2024 the orix developers
#
# This file is part of orix.
#
# orix is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# orix is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with orix.  If not, see <http://www.gnu.org/licenses/>.

"""Conversions of rotations between many common representations from
:cite:`rowenhorst2015consistent`, accelerated with Numba.

Conventions:

1. Right-handed Cartesian reference frames
2. Rotation angles are taken to be positive for a counter-clockwise
   rotation when viewing from the end point of the rotation axis unit
   vector towards the origin.
3. Rotations are *interpreted* in the passive sense. This means that we
   rotate reference frames with vectors fixed in space. Rotations are
   basis transformations rather than coordinate transformations.
4. Euler angle triplets are implemented using the Bunge convention, with
   angular ranges as [0, 2pi], [0, pi], and [0, 2pi].
5. Rotation angles are limited to [0, pi].
"""

from typing import Tuple

import numba as nb
import numpy as np

FLOAT_EPS = np.finfo(float).eps


@nb.jit("int64(float64[:])", cache=True, nogil=True, nopython=True)
def get_pyramid_single(xyz: np.ndarray) -> int:
    """Determine to which out of six pyramids in the cube a (x, y, z)
    coordinate belongs.

    Parameters
    ----------
    xyz
        1D array (x, y, z) of 64-bit floats.

    Returns
    -------
    pyramid
        Which pyramid ``xyz`` belongs to as a 64-bit integer.

    Notes
    -----
    This function is optimized with Numba, so care must be taken with
    array shapes and data types.
    """
    x, y, z = xyz
    x_abs, y_abs, z_abs = abs(x), abs(y), abs(z)
    if (x_abs <= z) and (y_abs <= z):  # Top
        return 1
    elif (x_abs <= -z) and (y_abs <= -z):  # Bottom
        return 2
    elif (z_abs <= x) and (y_abs <= x):  # Front
        return 3
    elif (z_abs <= -x) and (y_abs <= -x):  # Back
        return 4
    elif (x_abs <= y) and (z_abs <= y):  # Right
        return 5
    else:  # (x_abs <= -y) and (z_abs <= -y)  # Left
        return 6


@nb.jit("int64[:](float64[:, :])", cache=True, nogil=True, nopython=True)
def get_pyramid_2d(xyz: np.ndarray) -> np.ndarray:
    """Determine to which out of six pyramids in the cube a 2D array of
    (x, y, z) coordinates belongs.

    Parameters
    ----------
    xyz
        2D array of n (x, y, z) as 64-bit floats.

    Returns
    -------
    pyramids
        1D array of pyramids as 64-bit integers.

    Notes
    -----
    This function is optimized with Numba, so care must be taken with
    array shapes and data types.
    """
    n = xyz.shape[0]
    pyramids = np.zeros(n, dtype=np.int64)
    for i in nb.prange(n):
        pyramids[i] = get_pyramid_single(xyz[i])
    return pyramids


def get_pyramid(xyz: np.ndarray) -> np.ndarray:
    """N-dimensional wrapper for get_pyramid_2d, see the docstring of
    that function.
    """
    xyz2d = xyz.astype(np.float64)
    xyz2d = xyz2d.reshape(-1, 3)

    pyramids = get_pyramid_2d(xyz2d).ravel()

    return pyramids


@nb.jit("float64[:](float64[:])", cache=True, nogil=True, nopython=True)
def cu2ho_single(cu: np.ndarray) -> np.ndarray:
    """Convert a single set of cubochoric coordinates to un-normalized
    homochoric coordinates :cite:`singh2016orientation`.

    Parameters
    ----------
    cu
        1D array of (x, y, z) as 64-bit floats.

    Returns
    -------
    ho
        1D array of (x, y, z) as 64-bit floats.

    Notes
    -----
    This function is optimized with Numba, so care must be taken with
    array shapes and data types.
    """
    if np.max(np.abs(cu)) > (np.pi ** (2 / 3) / 2) + 1e-8:
        return np.zeros(3)

    # Determine which pyramid pair the point lies in and copy
    # coordinates in correct order
    pyramid = get_pyramid_single(cu)
    if pyramid in [1, 2]:
        pass
    elif pyramid in [3, 4]:
        cu = np.roll(cu, -1)
    else:  # [5, 6]
        cu = np.roll(cu, 1)

    # Scale by the grid parameter ratio
    cu = cu * np.pi ** (1 / 6) / 6 ** (1 / 6)

    cu_abs = np.abs(cu)
    if np.max(cu_abs) == 0:
        ho = np.zeros(3)
    else:
        if np.max(cu_abs[:2]) == 0:
            ho = np.array([0, 0, np.sqrt(6 / np.pi) * cu[2]])
        else:
            x, y, z = cu
            prefactor = (
                (3 * np.pi / 4) ** (1 / 3)
                * 2 ** (1 / 4)
                / (np.pi ** (5 / 6) / 6 ** (1 / 6) / 2)
            )
            sqrt2 = np.sqrt(2)
            if np.abs(y) <= np.abs(x):
                q = (np.pi / 12) * y / x
                cosq = np.cos(q)
                sinq = np.sin(q)
                q = prefactor * x / np.sqrt(sqrt2 - cosq)
                t1 = (sqrt2 * cosq - 1) * q
                t2 = sqrt2 * sinq * q
            else:
                q = (np.pi / 12) * x / y
                cosq = np.cos(q)
                sinq = np.sin(q)
                q = prefactor * y / np.sqrt(sqrt2 - cosq)
                t1 = sqrt2 * sinq * q
                t2 = (sqrt2 * cosq - 1) * q
            c = t1**2 + t2**2
            s = np.pi * c / (24 * z**2)
            c = np.sqrt(np.pi) * c / np.sqrt(24) / z
            q = np.sqrt(1 - s)

            ho = np.array([t1 * q, t2 * q, np.sqrt(6 / np.pi) * z - c])

    if pyramid in [1, 2]:
        return ho
    elif pyramid in [3, 4]:
        return np.roll(ho, 1)
    else:  # pyramid in [5, 6]
        return np.roll(ho, -1)


@nb.jit("float64[:, :](float64[:, :])", cache=True, nogil=True, nopython=True)
def cu2ho_2d(cu: np.ndarray) -> np.ndarray:
    """Convert multiple cubochoric coordinates to un-normalized
    homochoric coordinates :cite:`singh2016orientation`.

    Parameters
    ----------
    cu
        2D array of n (x, y, z) as 64-bit floats.

    Returns
    -------
    ho
        2D array of n (x, y, z) as 64-bit floats.

    Notes
    -----
    This function is optimized with Numba, so care must be taken with
    array shapes and data types.
    """
    ho = np.zeros_like(cu, dtype=np.float64)
    for i in nb.prange(cu.shape[0]):
        ho[i] = cu2ho_single(cu[i])
    return ho


def cu2ho(cu: np.ndarray) -> np.ndarray:
    """N-dimensional wrapper for cu2ho_2d, see the docstring of that
    function.
    """
    cu2d = cu.astype(np.float64)
    cu2d = cu2d.reshape(-1, 3)

    ho = cu2ho_2d(cu2d)
    ho = ho.reshape(cu.shape)

    return ho


@nb.jit("float64[:](float64[:])", cache=True, nogil=True, nopython=True)
def ho2ax_single(ho: np.ndarray) -> np.ndarray:
    """Convert a single set of homochoric coordinates to an
    un-normalized axis-angle pair :cite:`rowenhorst2015consistent`.

    Parameters
    ----------
    ho
        1D array of (x, y, z) as 64-bit floats.

    Returns
    -------
    ax
        1D array of (x, y, z, angle) as 64-bit floats.

    Notes
    -----
    This function is optimized with Numba, so care must be taken with
    array shapes and data types.
    """
    # Constants copied from EMsoft
    # fmt: off
    fit_parameters = np.array([
         0.9999999999999968,     -0.49999999999986866,     -0.025000000000632055,
        -0.003928571496460683,   -0.0008164666077062752,   -0.00019411896443261646,
        -0.00004985822229871769, -0.000014164962366386031, -1.9000248160936107e-6,
        -5.72184549898506e-6,     7.772149920658778e-6,    -0.00001053483452909705,
         9.528014229335313e-6,   -5.660288876265125e-6,     1.2844901692764126e-6,
         1.1255185726258763e-6,  -1.3834391419956455e-6,    7.513691751164847e-7,
        -2.401996891720091e-7,    4.386887017466388e-8,    -3.5917775353564864e-9
    ])
    # fmt: on
    ho_magnitude = np.sum(ho**2)
    if (ho_magnitude > -1e-8) and (ho_magnitude < 1e-8):
        ax = np.array([0, 0, 1, 0], dtype=np.float64)
    else:
        # Convert the magnitude to the rotation angle
        hom = ho_magnitude
        s = fit_parameters[0] + fit_parameters[1] * hom
        for i in nb.prange(2, 21):
            hom = hom * ho_magnitude
            s = s + fit_parameters[i] * hom
        hon = ho / np.sqrt(ho_magnitude)
        s = 2 * np.arccos(s)
        if np.abs(s - np.pi) < 1e-8:  # pragma: no cover
            ax = np.append(hon, np.pi)
        else:
            ax = np.append(hon, s)
    return ax


@nb.jit("float64[:, :](float64[:, :])", cache=True, nogil=True, nopython=True)
def ho2ax_2d(ho: np.ndarray) -> np.ndarray:
    """Convert multiple homochoric coordinates to un-normalized
    axis-angle pairs :cite:`rowenhorst2015consistent`.

    Parameters
    ----------
    ho
        2D array of n (x, y, z) as 64-bit floats.

    Returns
    -------
    ax
        2D array of n (x, y, z, angle) as 64-bit floats.

    Notes
    -----
    This function is optimized with Numba, so care must be taken with
    array shapes and data types.
    """
    n = ho.shape[0]
    ax = np.zeros((n, 4), dtype=np.float64)
    for i in nb.prange(n):
        ax[i] = ho2ax_single(ho[i])
    return ax


def ho2ax(ho: np.ndarray) -> np.ndarray:
    """N-dimensional wrapper for ho2ax_2d, see the docstring of that
    function.
    """
    ho2d = ho.astype(np.float64)
    ho2d = ho2d.reshape(-1, 3)

    ax = ho2ax_2d(ho2d)
    ax = ax.reshape(ho.shape[:-1] + (4,))

    return ax


@nb.jit("float64[:](float64[:])", cache=True, nogil=True, nopython=True)
def ax2ro_single(ax: np.ndarray) -> np.ndarray:
    """Convert a single angle-axis pair to an un-normalized Rodrigues
    vector :cite:`rowenhorst2015consistent`.

    Parameters
    ----------
    ax
        1D array of (x, y, z, angle) as 64-bit floats.

    Returns
    -------
    ro
        1D array of (x, y, z, angle) as 64-bit floats.

    Notes
    -----
    This function is optimized with Numba, so care must be taken with
    array shapes and data types.
    """
    ro = np.zeros(4, dtype=np.float64)
    angle = ax[3]
    if (angle > -1e-8) and (angle < 1e-8):
        ro[2] = 1
    else:
        ro[:3] = ax[:3]
        # Need to deal with the 180 degree case. A cutoff of 0.001 will
        # give a maximum Rodrigues magnitude of approximately 2000.
        # Raising it higher can cause rounding errors during conversions
        # to other rotation representations. If there is ever a
        # situation where this error must be smaller, the accuracy of
        # the test values for Homochoric and axis/angle in the tests
        # must be increased to more than 4 significant figures.
        if np.abs(angle - np.pi) < 1e-3:
            ro[3] = np.inf
        else:
            ro[3] = np.tan(angle * 0.5)
    return ro


@nb.jit("float64[:, :](float64[:, :])", cache=True, nogil=True, nopython=True)
def ax2ro_2d(ax: np.ndarray) -> np.ndarray:
    """Convert multiple axis-angle pairs to un-normalized Rodrigues
    vectors :cite:`rowenhorst2015consistent`.

    Parameters
    ----------
    ax
        2D array of n (x, y, z, angle) as 64-bit floats.

    Returns
    -------
    ro
        2D array of n (x, y, z, angle) as 64-bit floats.

    Notes
    -----
    This function is optimized with Numba, so care must be taken with
    array shapes and data types.
    """
    n = ax.shape[0]
    ro = np.zeros((n, 4), dtype=np.float64)
    for i in nb.prange(n):
        ro[i] = ax2ro_single(ax[i])
    return ro


def ax2ro(ax: np.ndarray) -> np.ndarray:
    """N-dimensional wrapper for ax2ro_2d, see the docstring of that
    function.
    """
    ax2d = ax.astype(np.float64)
    ax2d = ax2d.reshape(-1, 4)

    ro = ax2ro_2d(ax2d)
    ro = ro.reshape(ax.shape)

    return ro


@nb.jit("float64[:](float64[:])", cache=True, nogil=True, nopython=True)
def ro2ax_single(ro: np.ndarray) -> np.ndarray:
    """Convert a single Rodrigues vector to an un-normalized axis-angle
    pair :cite:`rowenhorst2015consistent`.

    Parameters
    ----------
    ro
        1D array of (x, y, z, angle) as 64-bit floats.

    Returns
    -------
    ax
        1D array of (x, y, z, angle) as 64-bit floats.

    Notes
    -----
    This function is optimized with Numba, so care must be taken with
    array shapes and data types.
    """
    if (ro[3] > -1e-8) and (ro[3] < 1e-8):
        return np.array([0, 0, 1, 0], dtype=np.float64)
    elif np.isinf(ro[3]):
        return np.append(ro[:3], np.pi)
    else:
        norm = np.sqrt(np.sum(np.square(ro[:3]), axis=-1))
        return np.append(ro[:3] / norm, 2 * np.arctan(ro[3]))


@nb.jit("float64[:, :](float64[:, :])", cache=True, nogil=True, nopython=True)
def ro2ax_2d(ro: np.ndarray) -> np.ndarray:
    """Convert multiple Rodrigues vectors to un-normalized axis-angle
    pairs :cite:`rowenhorst2015consistent`.

    Parameters
    ----------
    ro
        2D array of n (x, y, z, angle) as 64-bit floats.

    Returns
    -------
    ax
        2D array of n (x, y, z, angle) as 64-bit floats.

    Notes
    -----
    This function is optimized with Numba, so care must be taken with
    array shapes and data types.
    """
    n = ro.shape[0]
    ax = np.zeros((n, 4), dtype=np.float64)
    for i in nb.prange(n):
        ax[i] = ro2ax_single(ro[i])
    return ax


def ro2ax(ro: np.ndarray) -> np.ndarray:
    """N-dimensional wrapper for ro2ax_2d, see the docstring of that
    function.
    """
    ro2d = ro.astype(np.float64)
    ro2d = ro2d.reshape(-1, 4)

    ax = ro2ax_2d(ro2d)
    ax = ax.reshape(ro.shape)

    return ax


@nb.jit("float64[:](float64[:])", cache=True, nogil=True, nopython=True)
def ax2qu_single(ax: np.ndarray) -> np.ndarray:
    """Convert a single axis-angle pair to a unit quaternion
    :cite:`rowenhorst2015consistent`.

    Parameters
    ----------
    ax
        1D array of (x, y, z, angle) as 64-bit floats.

    Returns
    -------
    qu
        1D array of (a, b, c, d) as 64-bit floats.

    Notes
    -----
    This function is optimized with Numba, so care must be taken with
    array shapes and data types.
    """
    if (ax[3] > -1e-8) and (ax[3] < 1e-8):
        return np.array([1, 0, 0, 0], dtype=np.float64)
    else:
        c = np.cos(ax[3] * 0.5)
        s = np.sin(ax[3] * 0.5)
        qu = np.append(c, ax[:3] * s)
        norm = np.sqrt(np.sum(np.square(qu)))
        qu = qu / norm
        return qu


@nb.jit("float64[:, :](float64[:, :])", cache=True, nogil=True, nopython=True)
def ax2qu_2d(ax: np.ndarray) -> np.ndarray:
    """Convert multiple axis-angle pairs to unit quaternions
    :cite:`rowenhorst2015consistent`.

    Parameters
    ----------
    ax
        2D array of n (x, y, z, angle) as 64-bit floats.

    Returns
    -------
    qu
        2D array of n (a, b, c, d) as 64-bit floats.

    Notes
    -----
    This function is optimized with Numba, so care must be taken with
    array shapes and data types.
    """
    n = ax.shape[0]
    qu = np.zeros((n, 4), dtype=np.float64)
    for i in nb.prange(n):
        qu[i] = ax2qu_single(ax[i])
    return qu


def ax2qu(axes: np.ndarray, angles: np.ndarray) -> np.ndarray:
    """N-dimensional wrapper for ax2qu_2d, see the docstring of that
    function for further details.

    Parameters
    ----------
    axes
        N-dimensional array of (x, y, z) vectors with the final
        dimension equal to 3.
    angles
        Angles in radians.

    Returns
    -------
    qu
        2D array of n (a, b, c, d) as 64-bit floats.
    """
    axes = np.atleast_2d(axes)
    angles = np.atleast_1d(angles)

    if axes.shape[-1] != 3:
        raise ValueError("Final dimension of axes array must be 3.")
    if angles.shape[-1] != 1 or angles.shape == (1,) or axes.shape[:-1] == angles.shape:
        angles = angles.reshape(angles.shape + (1,))

    axes_shape = axes.shape[:-1]
    angles_shape = angles.shape[:-1]

    if angles_shape == (1,):
        # N-dimensional axis and single angle
        angles = np.ones(axes_shape + (1,)) * angles
    elif axes_shape == (1,):
        # Single axis and n-dimensional angle
        axes = np.ones(angles_shape + (3,)) * axes
    elif axes_shape != angles_shape:
        raise ValueError(
            f"The dimensions of axes {axes_shape} and angles {angles_shape} are "
            "incompatible. The dimensions must match or one must be a singular value."
        )

    axes_angles = np.concatenate([axes.data, angles], axis=-1)
    axes_angles_2d = axes_angles.reshape(-1, 4)
    axes_angles_2d = axes_angles_2d.astype(np.float64)

    qu = ax2qu_2d(axes_angles_2d)
    qu = qu.reshape(axes_angles.shape)

    return qu


@nb.jit("float64[:](float64[:])", cache=True, nogil=True, nopython=True)
def qu2ax_single(qu: np.ndarray) -> np.ndarray:
    """Convert a single (un)normalized quaternion to a normalized
    axis-angle pair :cite:`rowenhorst2015consistent`.

    Parameters
    ----------
    qu
        1D array of (a, b, c, d) as 64-bit floats.

    Returns
    -------
    ax
        1D array of (x, y, z, angle) as 64-bit floats.

    Notes
    -----
    Uses Eq. A.16 in :cite:`rowenhorst2015consistent`.

    This function is optimized with Numba, so care must be taken with
    array shapes and data types.
    """
    omega = 2 * np.arccos(qu[0])

    if omega < FLOAT_EPS:
        return np.array([0, 0, 1, 0], dtype=np.float64)

    if np.abs(qu[0]) < FLOAT_EPS:
        return np.array([qu[1], qu[2], qu[3], np.pi], dtype=np.float64)

    s = np.sqrt(np.sum(np.square(qu[1:])))
    if qu[0] <= 0:
        s = -s

    ax = np.array([qu[1], qu[2], qu[3], omega], dtype=np.float64)
    ax[:3] = ax[:3] / s

    return ax


@nb.jit("float64[:, :](float64[:, :])", cache=True, nogil=True, nopython=True)
def qu2ax_2d(qu: np.ndarray) -> np.ndarray:
    """Convert multiple (un)normalized quaternions to normalized
    axis-angle pairs :cite:`rowenhorst2015consistent`.

    Parameters
    ----------
    qu
        2D array of n (a, b, c, d) as 64-bit floats.

    Returns
    -------
    ax
        2D array of n (x, y, z, angle) as 64-bit floats.

    Notes
    -----
    This function is optimized with Numba, so care must be taken with
    array shapes and data types.
    """
    n = qu.shape[0]
    ax = np.zeros((n, 4), dtype=np.float64)
    for i in nb.prange(n):
        ax[i] = qu2ax_single(qu[i])
    return ax


def qu2ax(qu: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
    """N-dimensional wrapper for qu2ax_2d, see the docstring of that
    function for further details.

    Parameters
    ----------
    qu
        Quaternion(s) (a, b, c, d) with the final array dimension equal
        to 4.

    Returns
    -------
    axes
        Rotation axes of the same shape as the input array but with the
        final array dimension equal to 3.
    angles
        Rotation angles in radians of the same shape as the input array
        but with the final array dimension equal to 1.
    """
    qu_nd = np.atleast_2d(qu)

    if qu_nd.shape[-1] != 4:
        raise ValueError("Final dimension of quaternion array must be 4.")

    qu2d = qu.reshape(-1, 4)
    qu2d = qu2d.astype(np.float64)

    ax = qu2ax_2d(qu2d)
    ax = ax.reshape(qu.shape)

    axes = ax[..., :3]
    angles = ax[..., 3].reshape(ax.shape[:-1] + (1,))

    return axes, angles


@nb.jit("float64[:](float64[:])", cache=True, nogil=True, nopython=True)
def ho2ro_single(ho: np.ndarray) -> np.ndarray:
    """Convert a single set of homochoric coordinates to an
    un-normalized Rodrigues vector :cite:`rowenhorst2015consistent`.

    Parameters
    ----------
    ho
        1D array of (x, y, z) as 64-bit floats.

    Returns
    -------
    ro
        1D array of (x, y, z, angle) as 64-bit floats.

    Notes
    -----
    This function is optimized with Numba, so care must be taken with
    array shapes and data types.
    """
    return ax2ro_single(ho2ax_single(ho))


@nb.jit("float64[:, :](float64[:, :])", cache=True, nogil=True, nopython=True)
def ho2ro_2d(ho: np.ndarray) -> np.ndarray:
    """Convert multiple homochoric coordinates to un-normalized
    Rodrigues vectors :cite:`rowenhorst2015consistent`.

    Parameters
    ----------
    ho
        2D array of n (x, y, z) as 64-bit floats.

    Returns
    -------
    ax
        2D array of n (x, y, z, angle) as 64-bit floats.

    Notes
    -----
    This function is optimized with Numba, so care must be taken with
    array shapes and data types.
    """
    n = ho.shape[0]
    ro = np.zeros((n, 4), dtype=np.float64)
    for i in nb.prange(n):
        ro[i] = ho2ro_single(ho[i])
    return ro


def ho2ro(ho: np.ndarray) -> np.ndarray:
    """N-dimensional wrapper for ho2ro_2d, see the docstring of that
    function.
    """
    ho2d = ho.astype(np.float64)
    ho2d = ho2d.reshape(-1, 3)

    ro = ho2ro_2d(ho2d)
    ro = ro.reshape(ho.shape[:-1] + (4,))

    return ro


@nb.jit("float64[:](float64[:])", cache=True, nogil=True, nopython=True)
def cu2ro_single(cu: np.ndarray) -> np.ndarray:
    """Convert a single set of cubochoric coordinates to an
    un-normalized Rodrigues vector :cite:`rowenhorst2015consistent`.

    Parameters
    ----------
    cu
        1D array of (x, y, z) as 64-bit floats.

    Returns
    -------
    ro
        1D array of (x, y, z, angle) as 64-bit floats.

    Notes
    -----
    This function is optimized with Numba, so care must be taken with
    array shapes and data types.
    """
    if np.max(np.abs(cu)) == 0:
        return np.array([0, 0, 1, 0], dtype=np.float64)
    else:
        return ho2ro_single(cu2ho_single(cu))


@nb.jit("float64[:, :](float64[:, :])", cache=True, nogil=True, nopython=True)
def cu2ro_2d(cu: np.ndarray) -> np.ndarray:
    """Convert multiple cubochoric coordinates to un-normalized
    Rodrigues vectors :cite:`rowenhorst2015consistent`.

    Parameters
    ----------
    cu
        2D array of n (x, y, z) as 64-bit floats.

    Returns
    -------
    ro
        2D array of n (x, y, z, angle) as 64-bit floats.

    Notes
    -----
    This function is optimized with Numba, so care must be taken with
    array shapes and data types.
    """
    n = cu.shape[0]
    ro = np.zeros((n, 4), dtype=np.float64)
    for i in nb.prange(n):
        ro[i] = cu2ro_single(cu[i])
    return ro


def cu2ro(cu: np.ndarray) -> np.ndarray:
    """N-dimensional wrapper for cu2ro_2d, see the docstring of that
    function.
    """
    cu2d = cu.astype(np.float64)
    cu2d = cu2d.reshape(-1, 3)

    ro = cu2ro_2d(cu2d)
    ro = ro.reshape(cu.shape[:-1] + (4,))

    return ro


@nb.jit("float64[:](float64[:])", cache=True, nogil=True, nopython=True)
def eu2qu_single(eu: np.ndarray) -> np.ndarray:
    """Convert three Euler angles (alpha, beta, gamma) to a unit
    quaternion :cite:`rowenhorst2015consistent`.

    Parameters
    ----------
    eu
        1D array of (alpha, beta, gamma) Euler angles given in radians
        in the Bunge convention (i.e., passive Z-X-Z) as 64-bit floats.

    Returns
    -------
    qu
        1D unit quaternion (a, b, c, d) as 64-bit floats.

    Notes
    -----
    Uses Eqs. A.5 & A.6 in :cite:`rowenhorst2015consistent`.

    This function is optimized with Numba, so care must be taken with
    array shapes and data types.
    """
    sigma = 0.5 * np.add(eu[0], eu[2])
    delta = 0.5 * np.subtract(eu[0], eu[2])
    c = np.cos(eu[1] / 2)
    s = np.sin(eu[1] / 2)

    qu = np.zeros(4, dtype=np.float64)
    qu[0] = np.array(c * np.cos(sigma), dtype=np.float64)
    qu[1] = np.array(-s * np.cos(delta), dtype=np.float64)
    qu[2] = np.array(-s * np.sin(delta), dtype=np.float64)
    qu[3] = np.array(-c * np.sin(sigma), dtype=np.float64)

    if qu[0] < 0:
        qu = -qu

    return qu


@nb.jit("float64[:, :](float64[:, :])", cache=True, nogil=True, nopython=True)
def eu2qu_2d(eu: np.ndarray) -> np.ndarray:
    """Convert multiple Euler angles (alpha, beta, gamma) to unit
    quaternions.

    Parameters
    ----------
    eu
        2D array of n (alpha, beta, gamma) as 64-bit floats.

    Returns
    -------
    qu
        2D array of n (q0, q1, q2, q3) quaternions as 64-bit floats.

    Notes
    -----
    This function is optimized with Numba, so care must be taken with
    array shapes and data types.
    """
    n = eu.shape[0]
    qu = np.zeros((n, 4), dtype=np.float64)
    for i in nb.prange(n):
        qu[i] = eu2qu_single(eu[i])
    return qu


def eu2qu(eu: np.ndarray) -> np.ndarray:
    """N-dimensional wrapper for eu2qu_2d, see the docstring of that
    function.
    """
    eu2d = eu.astype(np.float64)
    eu2d = eu2d.reshape(-1, 3)

    qu = eu2qu_2d(eu2d)
    qu = qu.reshape(eu.shape[:-1] + (4,))

    return qu


@nb.jit("float64[:](float64[:, :])", cache=True, nogil=True, nopython=True)
def om2qu_single(om: np.ndarray) -> np.ndarray:
    """Convert a single (3, 3) rotation matrix to a unit quaternion
    :cite:`rowenhorst2015consistent`.

    Parameters
    ----------
    om
        (3, 3) rotation matrix as an array of 64-bit floats.

    Returns
    -------
    qu
        1D unit quaternion (a, b, c, d) as 64-bit floats.

    Notes
    -----
    Uses Eq. A.11 in :cite:`rowenhorst2015consistent`.

    This function is optimized with Numba, so care must be taken with
    array shapes and data types.
    """
    a_almost = 1 + om[0, 0] + om[1, 1] + om[2, 2]
    b_almost = 1 + om[0, 0] - om[1, 1] - om[2, 2]
    c_almost = 1 - om[0, 0] + om[1, 1] - om[2, 2]
    d_almost = 1 - om[0, 0] - om[1, 1] + om[2, 2]

    qu = np.zeros(4, dtype=np.float64)

    if a_almost < FLOAT_EPS:
        qu[0] = 0
    else:
        qu[0] = 0.5 * np.sqrt(a_almost)

    if b_almost < FLOAT_EPS:
        qu[1] = 0
    elif om[2, 1] < om[1, 2]:
        qu[1] = -0.5 * np.sqrt(b_almost)
    else:
        qu[1] = 0.5 * np.sqrt(b_almost)

    if c_almost < FLOAT_EPS:
        qu[2] = 0
    elif om[0, 2] < om[2, 0]:
        qu[2] = -0.5 * np.sqrt(c_almost)
    else:
        qu[2] = 0.5 * np.sqrt(c_almost)

    if d_almost < FLOAT_EPS:
        qu[3] = 0
    elif om[1, 0] < om[0, 1]:
        qu[3] = -0.5 * np.sqrt(d_almost)
    else:
        qu[3] = 0.5 * np.sqrt(d_almost)

    norm = np.sqrt(np.sum(np.square(qu)))
    qu = qu / norm

    return qu


@nb.jit("float64[:, :](float64[:, :, :])", cache=True, nogil=True, nopython=True)
def om2qu_3d(om: np.ndarray) -> np.ndarray:
    """Convert multiple rotation matrices to unit quaternions
    :cite:`rowenhorst2015consistent`.

    Parameters
    ----------
    om
        3D array of n (3, 3) rotation matrices as 64-bit floats.

    Returns
    -------
    qu
        2D array of n (a, b, c, d) quaternions as 64-bit floats.

    Notes
    -----
    This function is optimized with Numba, so care must be taken with
    array shapes and data types.
    """
    n = om.shape[0]
    qu = np.zeros((n, 4), dtype=np.float64)
    for i in nb.prange(n):
        qu[i] = om2qu_single(om[i])
    return qu


def om2qu(om: np.ndarray) -> np.ndarray:
    """N-dimensional wrapper for om2qu_3d, see the docstring of that
    function for further details.
    """
    om3d = om.reshape((-1, 3, 3))
    om3d = om3d.astype(np.float64)

    qu = om2qu_3d(om3d)
    qu = qu.reshape(om.shape[:-2] + (4,))

    return qu


@nb.jit("float64[:](float64[:])", cache=True, nogil=True, nopython=True)
def qu2eu_single(qu: np.ndarray) -> np.ndarray:
    """Convert a unit quaternion to three Euler angles
    :cite:`rowenhorst2015consistent`.

    Parameters
    ----------
    qu
        Unit quaternion (a, b, c, d).

    Return
    ------
    eu
        Euler angles (alpha, beta, gamma) in radians in the Bunge
        convention (i.e., passive Z-X-Z).

    Notes
    -----
    Uses Eq. A.14 in :cite:`rowenhorst2015consistent`.

    This function is optimized with Numba, so care must be taken with
    array shapes and data types.
    """
    eu = np.zeros(3, dtype=np.float64)

    # Following Eq. A.14, compute q_ad, q_bc, and chi, assuming P=1
    q_ad = (qu[0] * qu[0]) + (qu[3] * qu[3])
    q_bc = (qu[1] * qu[1]) + (qu[2] * qu[2])
    chi = np.sqrt(q_ad * q_bc)

    if chi < FLOAT_EPS:
        if q_bc < FLOAT_EPS:
            a = -2 * qu[0] * qu[3]
            b = qu[0] * qu[0] - qu[3] * qu[3]
        else:
            a = -2 * qu[1] * qu[2]
            b = qu[1] * qu[1] - qu[2] * qu[2]
            eu[1] = np.pi
        eu[0] = np.arctan2(a, b)
        return np.mod(eu, np.pi * 2)

    eu_0a = (qu[1] * qu[3] - qu[0] * qu[2]) / chi
    eu_0b = (-qu[0] * qu[1] - qu[2] * qu[3]) / chi
    eu_2a = (qu[0] * qu[2] + qu[1] * qu[3]) / chi
    eu_2b = (qu[2] * qu[3] - qu[0] * qu[1]) / chi

    eu[0] = np.arctan2(eu_0a, eu_0b)
    eu[1] = np.arctan2(2 * chi, q_ad - q_bc)
    eu[2] = np.arctan2(eu_2a, eu_2b)

    eu[np.abs(eu) < FLOAT_EPS] = 0

    return np.mod(eu, np.pi * 2)


@nb.jit("float64[:, :](float64[:, :])", cache=True, nogil=True, nopython=True)
def qu2eu_2d(qu: np.ndarray) -> np.ndarray:
    """Convert multiple unit quaternions to Euler angles.

    Parameters
    ----------
    qu
        2D array of n (a, b, c, d) quaternions as 64-bit floats.

    Returns
    -------
    eu
        2D array of n (alpha, beta, gamma) as 64-bit floats.

    Notes
    -----
    This function is optimized with Numba, so care must be taken with
    array shapes and data types.
    """
    n = qu.shape[0]
    eu = np.zeros((n, 3), dtype=np.float64)
    for i in nb.prange(n):
        eu[i] = qu2eu_single(qu[i])
    return eu


def qu2eu(qu: np.ndarray) -> np.ndarray:
    """N-dimensional wrapper for qu2eu_2d, see the docstring of that
    function.
    """
    qu2d = qu.reshape(-1, 4)
    qu2d = qu2d.astype(np.float64)

    eu = qu2eu_2d(qu2d)
    eu = eu.reshape(qu.shape[:-1] + (3,))

    return eu


@nb.jit("float64[:, :](float64[:])", cache=True, nogil=True, nopython=True)
def qu2om_single(qu: np.ndarray) -> np.ndarray:
    """Convert a unit quaternion to an orthogonal rotation matrix
     :cite:`rowenhorst2015consistent`.

    Parameters
    ----------
    qu
        1D unit quaternion (a, b, c, d) as 64-bit floats.

    Returns
    -------
    om
        (3, 3) rotation matrix as an array of 64-bit floats.

    Notes
    -----
    Uses Eq. A.15 :cite:`rowenhorst2015consistent`.

    This function is optimized with Numba, so care must be taken with
    array shapes and data types.
    """
    om = np.zeros((3, 3), dtype=np.float64)

    bb = qu[1] ** 2
    cc = qu[2] ** 2
    dd = qu[3] ** 2
    qq = qu[0] ** 2 - (bb + cc + dd)  # q_mean in Eq. A.15

    bc = qu[1] * qu[2]
    ad = qu[0] * qu[3]
    bd = qu[1] * qu[3]
    ac = qu[0] * qu[2]
    cd = qu[2] * qu[3]
    ab = qu[0] * qu[1]
    om[0, 0] = qq + 2 * bb
    om[0, 1] = 2 * (bc - ad)
    om[0, 2] = 2 * (bd + ac)
    om[1, 0] = 2 * (bc + ad)
    om[1, 1] = qq + 2 * cc
    om[1, 2] = 2 * (cd - ab)
    om[2, 0] = 2 * (bd - ac)
    om[2, 1] = 2 * (cd + ab)
    om[2, 2] = qq + 2 * dd

    return om


@nb.jit("float64[:, :, :](float64[:, :])", cache=True, nogil=True, nopython=True)
def qu2om_2d(qu: np.ndarray) -> np.ndarray:
    """Convert multiple unit quaternions to orthogonal rotation
    matrices.

    Parameters
    ----------
    qu
        2D array of n (q0, q1, q2, q3) quaternions as 64-bit floats.

    Returns
    -------
    om
        3D array of n (3, 3) rotation matrices as 64-bit floats.

    Notes
    -----
    This function is optimized with Numba, so care must be taken with
    array shapes and data types.
    """
    n = qu.shape[0]
    om = np.zeros((n, 3, 3), dtype=np.float64)
    for i in nb.prange(n):
        om[i, :, :] = qu2om_single(qu[i])
    return om


def qu2om(qu: np.ndarray) -> np.ndarray:
    """N-dimensional wrapper for om2qu_3d, see the docstring of that
    function.
    """
    qu2d = qu.reshape(-1, 4)
    qu2d = qu2d.astype(np.float64)

    om = qu2om_2d(qu2d)
    om = om.reshape(qu.shape[:-1] + (3, 3))

    return om


@nb.jit("float64[:](float64[:])", cache=True, nogil=True, nopython=True)
def qu2ho_single(qu: np.ndarray) -> np.ndarray:
    """Convert a single (un)normalized quaternion to a normalized
    homochoric vector :cite:`rowenhorst2015consistent`.

    Parameters
    ----------
    qu
        1D array of (a, b, c, d) as 64-bit floats.

    Returns
    -------
    ho
        1D array of (x, y, z) as 64-bit floats.

    Notes
    -----
    Uses Eq. A.25 in :cite:`rowenhorst2015consistent`.

    This function is optimized with Numba, so care must be taken with
    array shapes and data types.
    """
    omega = 2 * np.arccos(qu[0])

    if omega < FLOAT_EPS:
        return np.zeros(3, dtype=np.float64)

    s = np.sqrt(np.sum(np.square(qu[1:])))
    n = qu[1:] / s
    f = 3 * (omega - np.sin(omega)) / 4
    ho = n * f ** (1 / 3)

    return ho


@nb.jit("float64[:, :](float64[:, :])", cache=True, nogil=True, nopython=True)
def qu2ho_2d(qu: np.ndarray) -> np.ndarray:
    """Convert multiple (un)normalized quaternions to normalized
    homochoric vectors :cite:`rowenhorst2015consistent`.

    Parameters
    ----------
    qu
        2D array of n (a, b, c, d) as 64-bit floats.

    Returns
    -------
    ho
        2D array of n (x, y, z) as 64-bit floats.

    Notes
    -----
    This function is optimized with Numba, so care must be taken with
    array shapes and data types.
    """
    n = qu.shape[0]
    ho = np.zeros((n, 3), dtype=np.float64)
    for i in nb.prange(n):
        ho[i] = qu2ho_single(qu[i])
    return ho


def qu2ho(qu: np.ndarray) -> np.ndarray:
    """N-dimensional wrapper for qu2ho_2d, see the docstring of that
    function for further details.

    Parameters
    ----------
    qu
        Quaternion(s) (a, b, c, d) with the final array dimension equal
        to 4.

    Returns
    -------
    ho
        Homochoric vectors (x, y, z).
    """
    qu_nd = np.atleast_2d(qu)

    if qu_nd.shape[-1] != 4:
        raise ValueError("Final dimension of quaternion array must be 4.")

    qu2d = qu.reshape(-1, 4)
    qu2d = qu2d.astype(np.float64)

    ho = qu2ho_2d(qu2d)
    ho = ho.reshape(qu.shape[:-1] + (3,))

    return ho



# -*- coding: utf-8 -*-
# Copyright 2018-2024 the orix developers
#
# This file is part of orix.
#
# orix is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# orix is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with orix.  If not, see <http://www.gnu.org/licenses/>.

from typing import Optional, Tuple, Union

from matplotlib import projections
from matplotlib.gridspec import SubplotSpec
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D

from orix.vector import AxAngle, Rodrigues


class RotationPlot(Axes3D):
    """Plot of a (mis)orientation region."""

    name = None
    transformation_class = None

    def transform(
        self,
        xs: Union["Misorientation", "OrientationRegion", "Rotation"],
        fundamental_zone: Optional["OrientationRegion"] = None,
    ):
        """Prepare (mis)orientations or rotations for plotting.

        Parameters
        ----------
        xs
            Object to transform.
        fundamental_zone
            Orientation region to add to the plot as a wireframe.
        """
        from orix.quaternion import Misorientation, OrientationRegion, Rotation

        # Project rotations into fundamental zone if necessary
        if isinstance(xs, Misorientation):
            if fundamental_zone is None:
                symmetry = xs.symmetry
                # Orientation.symmetry returns a Symmetry object not a tuple, so pack
                if not isinstance(symmetry, tuple):
                    symmetry = (symmetry,)
                fundamental_zone = OrientationRegion.from_symmetry(*symmetry)
            # check fundamental_zone is properly defined
            if not isinstance(fundamental_zone, OrientationRegion):
                raise TypeError("fundamental_zone is not an OrientationRegion object.")
            # if any in xs are out of fundamental_zone, calculate symmetry reduction
            if not (xs < fundamental_zone).all():
                xs = xs.map_into_symmetry_reduced_zone()

        if isinstance(xs, Rotation):
            if isinstance(xs, OrientationRegion):
                xs = xs.get_plot_data()
            transformed = self.transformation_class.from_rotation(xs)
        else:
            transformed = self.transformation_class(xs)
        x, y, z = transformed.xyz
        return x, y, z

    def scatter(
        self, xs: Union["Misorientation", "Rotation"], fundamental_zone=None, **kwargs
    ):
        """Create a scatter plot.

        Parameters
        ----------
        xs
            Rotations.
        fundamental_zone
            Orientation region to add to the plot as a wireframe.
        **kwargs
            Keyword arguments passed to
            :meth:`mpl_toolkits.mplot3d.Axes3D.scatter`.
        """
        x, y, z = self.transform(xs, fundamental_zone=fundamental_zone)
        return super().scatter(x, y, z, **kwargs)

    def plot(self, xs, **kwargs):
        x, y, z = self.transform(xs)
        return super().plot(x, y, z, **kwargs)

    def plot_wireframe(self, xs, **kwargs):
        d = dict(color="gray", alpha=0.5, linewidth=0.5, rcount=30, ccount=30)
        for k, v in d.items():
            kwargs.setdefault(k, v)
        x, y, z = self.transform(xs)
        return super().plot_wireframe(x, y, z, **kwargs)

    def _get_region_extent(
        self, fundamental_region: "OrientationRegion"
    ) -> Tuple[float, float, float]:
        """Return the maximum angles in x, y, z of the fundamental
        region.

        Parameters
        ----------
        fundamental_region

        Returns
        -------
        tuple of float
        """
        x, y, z = self.transform(fundamental_region)
        return x.max(), y.max(), z.max()

    def _correct_aspect_ratio(self, fundamental_region, set_limits=True):
        """Correct the aspect ratio of the axis according to the
        extent of the boundaries of the fundamental region.

        Parameters
        ----------
        fundamental_region : OrientationRegion
        set_limits : bool, optional
            Whether to also restrict the data limits to the boundary
            extents. Default is True.
        """
        xlim, ylim, zlim = self._get_region_extent(fundamental_region)
        self.set_box_aspect((xlim, ylim, zlim))
        if set_limits:
            self.set_xlim(-xlim, xlim)
            self.set_ylim(-ylim, ylim)
            self.set_zlim(-zlim, zlim)


class RodriguesPlot(RotationPlot):
    """Plot rotations in Rodrigues-Frank space."""

    name = "rodrigues"
    transformation_class = Rodrigues


class AxAnglePlot(RotationPlot):
    """Plot rotations in a axis-angle space."""

    name = "axangle"
    transformation_class = AxAngle


projections.register_projection(RodriguesPlot)
projections.register_projection(AxAnglePlot)


def _setup_rotation_plot(
    figure: Optional[plt.Figure] = None,
    projection: str = "axangle",
    position: Union[int, tuple, SubplotSpec, None] = (1, 1, 1),
    figure_kwargs: Optional[dict] = None,
) -> Tuple[plt.Figure, Union[AxAnglePlot, RodriguesPlot]]:
    """Return a figure and rotation plot axis of the correct type.

    This is a convenience method used in e.g.
    :meth:`orix.quaternion.Orientation.scatter`.

    Parameters
    ----------
    figure
        If given, a new plot axis :class:`orix.plot.AxAnglePlot` or
        :class:`orix.plot.RodriguesPlot` is added to the figure in
        the position specified by `position`. If not given, a new
        figure is created.
    projection
        Which orientation space to plot orientations in, either
        "axangle" (default) or "rodrigues".
    position
        Where to add the new plot axis. 121 or (1, 2, 1) places it
        in the first of two positions in a grid of 1 row and 2
        columns. See :meth:`matplotlib.figure.Figure.add_subplot`
        for further details. Default is (1, 1, 1).
    figure_kwargs
        Dictionary of keyword arguments passed to
        :func:`matplotlib.pyplot.figure` if `figure` is not given.

    Returns
    -------
    figure
        Figure with the added plot axis.
    ax
        The axis-angle or Rodrigues plot axis.
    """
    if figure is None:
        if figure_kwargs is None:
            figure_kwargs = {"layout": "tight"}
        figure = plt.figure(**figure_kwargs)

    subplot_kwds = {
        "projection": projection,
        "proj_type": "ortho",
        "auto_add_to_figure": False,
    }

    if hasattr(position, "__iter__"):
        # (nrows, ncols, index)
        ax = figure.add_subplot(*position, **subplot_kwds)
    else:
        # Position, e.g. 121, or `SubplotSpec`
        ax = figure.add_subplot(position, **subplot_kwds)

    return figure, ax



# -*- coding: utf-8 -*-
# Copyright 2018-2024 the orix developers
#
# This file is part of orix.
#
# orix is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# orix is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with orix.  If not, see <http://www.gnu.org/licenses/>.

from typing import Tuple, Union

from matplotlib.patches import FancyArrowPatch
from mpl_toolkits.mplot3d import proj3d
import numpy as np


class Arrow3D(FancyArrowPatch):
    """Matplotlib arrows in 3D with arrowheads.

    Initially taken from this Stack Overflow answer by user CT Zhu
    https://stackoverflow.com/a/22867877/12063126.
    """

    def __init__(self, xs, ys, zs, *args, **kwargs):
        super().__init__((0, 0), (0, 0), *args, **kwargs)
        self._verts3d = xs, ys, zs

    def do_3d_projection(self, renderer=None):
        xs3d, ys3d, zs3d = self._verts3d
        xs, ys, zs = proj3d.proj_transform(xs3d, ys3d, zs3d, self.axes.M)
        self.set_positions((xs[0], ys[0]), (xs[1], ys[1]))
        return np.min(zs)


def format_labels(
    v: Union[list, tuple, np.ndarray],
    brackets: Tuple[str, str] = ("", ""),
    use_latex: bool = True,
) -> np.ndarray:
    r"""Return formatted vector integer labels.

    This function is a convenient way to get nice labels when plotting
    vectors in the stereographic projection via
    :meth:`~orix.vector.Vector3d.scatter` or
    :meth:`~orix.plot.StereographicPlot.text`.

    Parameters
    ----------
    v
        Vector labels in an array-like with the last dimension having a
        size of 3 or 4. The labels are rounded to the closets integers
        before being formatted.
    brackets
        Left and right parentheses. These are typically () or [] when
        labeling reciprocal (hkl) or direct [uvw] lattice vectors,
        respectively. If not given, no parentheses are added. If
        ``use_latex=True``, then the brackets {} and <> will be escaped
        if given.
    use_latex
        Whether to use LaTeX when formatting the labels. Default is
        True.

    Returns
    -------
    new_labels
        Array of string labels.

    Examples
    --------
    >>> from orix import plot
    >>> from orix.vector import Vector3d
    >>> v = Vector3d([[1, 1, 1], [-2, 0, 1], [4, 0, 0], [-4, 0, 0]])
    >>> plot.format_labels(v.reshape(2, 2).data)
    array([['$111$', '$\\bar{2}01$'],
           ['$400$', '$\\bar{4}00$']], dtype='<U11')
    >>> plot.format_labels(v.data, ("[", "]"), use_latex=False).tolist()
    ['[111]', '[-201]', '[400]', '[-400]']
    >>> plot.format_labels(v.data, ("{", "}")).tolist()
    ['$\\{111\\}$', '$\\{\\bar{2}01\\}$', '$\\{400\\}$', '$\\{\\bar{4}00\\}$']
    >>> plot.format_labels(v.data, ("{", "}"), use_latex=False).tolist()
    ['{111}', '{-201}', '{400}', '{-400}']
    >>> plot.format_labels(v.data, ("<", ">")).tolist()
    ['$\\left<111\\right>$',
     '$\\left<\\bar{2}01\\right>$',
     '$\\left<400\\right>$',
     '$\\left<\\bar{4}00\\right>$']
    """
    if use_latex:
        start = end = r"$"
        if brackets == ("<", ">"):
            brackets = (r"\left<", r"\right>")
        elif brackets == ("{", "}"):
            brackets = (r"\{", r"\}")
    else:
        start = end = ""

    v = np.asanyarray(v)
    shape = v.shape
    v = v.round().astype(int).reshape(-1, shape[-1])

    new_labels = []
    for label in v:
        new_label = start + brackets[0]
        for i in label:
            if i < 0 and use_latex:
                new_label += r"\bar{" + str(abs(i)) + r"}"
            else:
                new_label += str(i)
        new_label += brackets[1] + end
        new_labels.append(new_label)

    return np.asarray(new_labels).reshape(shape[:-1])



# -*- coding: utf-8 -*-
# Copyright 2018-2024 the orix developers
#
# This file is part of orix.
#
# orix is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# orix is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with orix.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np
import pytest

from orix.quaternion import Rotation
from orix.quaternion.symmetry import C2, C4, C6, D6, Oh, get_point_group
from orix.sampling import (
    get_sample_fundamental,
    get_sample_local,
    get_sample_reduced_fundamental,
    uniform_SO3_sample,
)
from orix.sampling.SO3_sampling import _resolution_to_num_steps
from orix.sampling._polyhedral_sampling import (
    _get_angles_between_nn_gridpoints,
    _get_first_nearest_neighbors,
    _get_max_grid_angle,
    _get_start_and_end_index,
)
from orix.vector import Vector3d


@pytest.fixture(scope="session")
def sample():
    return uniform_SO3_sample(2, method="haar_euler")


@pytest.fixture(scope="session")
def fixed_rotation():
    """A fixed rotation."""
    return Rotation([0.5, 0.5, 0, 0])


class TestPolyhedralSamplingUtils:
    @pytest.mark.parametrize(
        "number_of_steps, include_start, include_end, positive_and_negative, expected",
        [
            (5, False, False, False, (1, 5)),
            (5, True, False, False, (0, 5)),
            (6, False, True, False, (1, 7)),
            (7, True, True, True, (-7, 8)),
        ],
    )
    def test_get_start_and_end_index(
        self,
        number_of_steps,
        include_start,
        include_end,
        positive_and_negative,
        expected,
    ):
        assert (
            _get_start_and_end_index(
                number_of_steps=number_of_steps,
                include_start=include_start,
                include_end=include_end,
                positive_and_negative=positive_and_negative,
            )
            == expected
        )

    def test_first_nearest_neighbors(self):
        grid = np.array(
            [
                [1, 0, 0],
                [0, 1, 0],
                [0, 1, 1],
                [1, 0, 1],
            ]
        )
        fnn = np.array(
            [
                [1, 0, 1],
                [0, 1, 1],
                [0, 1, 0],
                [1, 0, 0],
            ]
        )
        angles = np.array([45, 45, 45, 45])
        fnn_test = _get_first_nearest_neighbors(grid)
        angles_test = _get_angles_between_nn_gridpoints(grid)
        assert np.allclose(fnn, fnn_test)
        assert np.allclose(angles, angles_test)
        assert abs(_get_max_grid_angle(grid) - 45.0) < 1e-7


class TestSamplingUtils:
    def test_resolution_to_num_steps(self):
        n_steps_odd = _resolution_to_num_steps(1, odd_only=True)
        assert isinstance(n_steps_odd, int)
        assert n_steps_odd == 361
        # Note that 360/1.002 = 359.28 rounds to 359
        assert _resolution_to_num_steps(1.002, even_only=True) == 360


class TestUniformSO3:
    def test_sample_size(self):
        r1 = uniform_SO3_sample(10, method="haar_euler")
        assert r1.size > 0
        r2 = uniform_SO3_sample(10, method="quaternion")
        assert r2.size > 0

    def test_uniform_SO3_sample_regions(self, sample, fixed_rotation):
        """Checks that different regions have the same density."""
        around_zero = sample[sample.a > 0.9]
        moved = fixed_rotation * sample
        elsewhere = moved[sample.a > 0.9]
        # Extra line simplifies the stack trace
        x, y = around_zero.size, elsewhere.size
        assert np.isclose(x, y, rtol=0.025)

    def test_uniform_SO3_sample_resolution(self, sample):
        """Checks that doubling resolution doubles density (8-fold
        counts).
        """
        lower = uniform_SO3_sample(4, method="haar_euler")
        x, y = lower.size * 8, sample.size
        assert np.isclose(x, y, rtol=0.025)


class TestGetSampleLocal:
    @pytest.mark.parametrize(
        "big, small, method", [(180, 90, "quaternion"), (180, 90, "haar_euler")]
    )
    def test_get_sample_local_width(self, big, small, method):
        """Checks that size follows the expected trend (X - Sin(X)).
        With 180 and 90 we expect: pi and pi/2 - 1.
        """
        res = 4  # Degrees
        x_size = get_sample_local(resolution=res, grid_width=small, method=method).size
        y_size = get_sample_local(resolution=res, grid_width=big, method=method).size
        x_v = np.deg2rad(small) - np.sin(np.deg2rad(small))
        y_v = np.deg2rad(big) - np.sin(np.deg2rad(big))
        exp = y_size / x_size
        theory = y_v / x_v

        assert x_size > 0  # If this fails exp will be NaN
        assert np.isclose(exp, theory, atol=0.2)

    def test_get_sample_local_center(self, fixed_rotation):
        # fixed rotation takes us 30 degrees from origin
        x = get_sample_local(
            resolution=3, grid_width=20, center=fixed_rotation, method="haar_euler"
        )
        # a == cos(omega/2)
        assert np.all(x.a < np.cos(np.deg2rad(5)))


class TestSampleFundamental:
    @pytest.fixture(scope="session")
    def C6_sample(self):
        return get_sample_fundamental(resolution=4, point_group=C6, method="haar_euler")

    def test_get_sample_fundamental_zone_order(self, C6_sample):
        """Cross check point counts to group order terms."""
        D6_sample = get_sample_fundamental(4, point_group=D6, method="haar_euler")
        ratio = C6_sample.size / D6_sample.size
        assert np.isclose(ratio, 2, atol=0.2)

    def test_get_sample_fundamental_space_group(self, C6_sample):
        """Going via the `space_group` route."""
        # Assert that space group #3 is has pg C2
        assert C2 == get_point_group(3, proper=True)
        C2_sample = get_sample_fundamental(4, space_group=3, method="haar_euler")
        ratio = C2_sample.size / C6_sample.size
        assert np.isclose(ratio, 3, atol=0.2)

    def test_get_sample_reduced_fundamental(self):
        vz = Vector3d.zvector()

        R_C1 = get_sample_reduced_fundamental(resolution=4)
        v_C1 = R_C1 * vz
        assert np.allclose(v_C1.mean().data, [0, 0, 0])

        R_C4 = get_sample_reduced_fundamental(resolution=4, point_group=C4)
        v_C4 = R_C4 * vz
        assert np.all(v_C4 <= C4.fundamental_sector)

        R_C6 = get_sample_reduced_fundamental(resolution=4, point_group=C6)
        v_C6 = R_C6 * vz
        assert np.all(v_C6 <= C6.fundamental_sector)

        R_Oh = get_sample_reduced_fundamental(point_group=Oh)
        v_Oh = R_Oh * vz
        assert np.all(v_Oh <= Oh.fundamental_sector)

        # Some rotations have a phi1 Euler angle of multiples of pi,
        # presumably due to rounding errors
        phi1_C1 = R_C1.to_euler()[:, 0].round(7)
        assert np.allclose(np.unique(phi1_C1), [0, 2 * np.pi], atol=1e-7)
        phi1_C4 = R_C4.to_euler()[:, 0].round(7)
        assert np.allclose(np.unique(phi1_C4), [0, np.pi / 2], atol=1e-7)
        phi1_C6 = R_C6.to_euler()[:, 0].round(7)
        assert np.allclose(np.unique(phi1_C6), [0, np.pi / 2], atol=1e-7)
        phi1_Oh = R_Oh.to_euler()[:, 0].round(7)
        assert np.allclose(np.unique(phi1_Oh), [0, np.pi / 2], atol=1e-7)

        R_Oh2 = get_sample_reduced_fundamental(point_group=Oh, method="icosahedral")
        assert R_Oh.size > R_Oh2.size



# -*- coding: utf-8 -*-
# Copyright 2018-2024 the orix developers
#
# This file is part of orix.
#
# orix is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# orix is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with orix.  If not, see <http://www.gnu.org/licenses/>.

"""Base class for three-dimensional objects."""

from __future__ import annotations

from typing import Any, Optional, Tuple, Union

import numpy as np


class DimensionError(Exception):
    """Error raised when an array passed to a class constructor has an
    incompatible shape.

    Parameters
    ----------
    this
        An orix class with attributes ``dim`` and ``shape``.
    data
        Array.
    """

    def __init__(self, this: Object3d, data: np.ndarray):
        super().__init__(
            f"{this.__class__.__name__} requires data of dimension {this.dim}, but "
            f"received dimension {data.shape[-1]}"
        )


class Object3d:
    """Base class for 3d objects.

    .. note::

        This class is not meant to be used directly.

    Parameters
    ----------
    data
        Object data.
    """

    dim = None
    """Return the number of dimensions for this object."""  # pragma: no cover

    _data = None

    __array_ufunc__ = None

    def __init__(self, data=None):
        if isinstance(data, Object3d):
            self._data = data.data
        else:
            data = np.atleast_2d(data)
            if data.shape[-1] != self.dim:
                raise DimensionError(self, data)
            self._data = data
        self.__finalize__(data)

    def __finalize__(self, data):
        pass

    # -------------------------- Properties -------------------------- #

    @property
    def data(self) -> np.ndarray:
        """Return the data."""
        return self._data[..., : self.dim]

    @data.setter
    def data(self, data: np.ndarray):
        """Set the data."""
        self._data[..., : self.dim] = data

    @property
    def shape(self) -> tuple:
        """Return the shape of the object."""
        return self.data.shape[:-1]

    @property
    def ndim(self) -> int:
        """Return the number of navigation dimensions of the object.

        For example, if :attr:`data` has shape (4, 5, 6), :attr:`ndim`
        is 3.
        """
        return len(self.shape)

    @property
    def size(self) -> int:
        """Return the total number of entries in this object."""
        return int(np.prod(self.shape))

    @property
    def norm(self) -> np.ndarray:
        """Return the norm of the data."""
        return np.sqrt(np.sum(np.square(self.data), axis=-1))

    @property
    def unit(self) -> Object3d:
        """Return the unit object."""
        with np.errstate(divide="ignore", invalid="ignore"):
            obj = self.__class__(np.nan_to_num(self.data / self.norm[..., np.newaxis]))
            return obj

    # ------------------------ Dunder methods ------------------------ #

    def __repr__(self) -> str:
        """Return a string representation of the data."""
        name = self.__class__.__name__
        shape = str(self.shape)
        data = np.array_str(self.data, precision=4, suppress_small=True)
        return "\n".join([name + " " + shape, data])

    def __getitem__(self, key) -> Object3d:
        """Return a slice of the object."""
        data = np.atleast_2d(self.data[key])
        obj = self.__class__(data)
        return obj

    def __setitem__(self, key, value: np.ndarray):
        """Set a slice of the data."""
        self.data[key] = value.data

    # ------------------------ Class methods ------------------------- #

    @classmethod
    def empty(cls) -> Object3d:
        """Return an empty object with the appropriate dimensions."""
        return cls(np.zeros((0, cls.dim)))

    @classmethod
    def stack(cls, sequence: Any) -> Object3d:
        """Return a stacked object from the sequence.

        Parameters
        ----------
        sequence
            A sequence of objects to stack.
        """
        sequence = [s._data for s in sequence]
        stack = np.stack(sequence, axis=-2)
        obj = cls(stack[..., : cls.dim])
        obj._data = stack
        return obj

    @classmethod
    def random(cls, shape: Union[int, tuple] = 1) -> Object3d:
        """Create object with random data.

        Parameters
        ----------
        shape
            Shape of the object.

        Returns
        -------
        obj
            Object with random data.
        """
        n = int(np.prod(shape))
        obj = []
        while len(obj) < n:
            r = np.random.uniform(-1, 1, (3 * n, cls.dim))
            r2 = np.sum(np.square(r), axis=1)
            r = r[np.logical_and(1e-9**2 < r2, r2 <= 1)]
            obj += list(r)
        obj = cls(np.array(obj[:n]))
        obj = obj.unit
        obj = obj.reshape(shape)
        return obj

    # --------------------- Other public methods --------------------- #

    def flatten(self):
        """Return a new object with the same data in a single column."""
        obj = self.__class__(self.data.T.reshape(self.dim, -1).T)
        real_dim = self._data.shape[-1]
        obj._data = self._data.T.reshape(real_dim, -1).T
        return obj

    def unique(self, return_index: bool = False, return_inverse: bool = False) -> Union[
        Tuple[Object3d, np.ndarray, np.ndarray],
        Tuple[Object3d, np.ndarray],
        Object3d,
    ]:
        """Return a new object containing only this object's unique
        entries.

        Unless overridden, this method returns the numerically unique
        entries. It also removes zero-entries which are assumed to be
        degenerate.

        Parameters
        ----------
        return_index
            If ``True``, will also return the indices of the (flattened)
            data where the unique entries were found.
        return_inverse
            If ``True``, will also return the indices to reconstruct the
            (flattened) data from the unique data.

        Returns
        -------
        dat
            The numerically unique entries.
        idx
            The indices of the unique data in the (flattened) array if
            ``return_index=True``.
        inv
            The indices of the (flattened) data in the unique array if
            ``return_inverse=True``.
        """
        data = self.flatten()._data.round(10)
        data = data[~np.all(np.isclose(data, 0), axis=1)]  # Remove zeros
        _, idx, inv = np.unique(data, axis=0, return_index=True, return_inverse=True)
        obj = self.__class__(data[np.sort(idx), : self.dim])
        obj._data = data[np.sort(idx)]
        if return_index and return_inverse:
            return obj, idx, inv
        elif return_index and not return_inverse:
            return obj, idx
        elif return_inverse and not return_index:
            return obj, inv
        else:
            return obj

    def squeeze(self) -> Object3d:
        """Return a new object with the same data with length
        1-dimensions removed.

        Returns
        -------
        obj
            Squeezed object.
        """
        obj = self.__class__(self)
        obj._data = np.atleast_2d(self._data.squeeze())
        return obj

    def reshape(self, *shape: Union[int, tuple]) -> Object3d:
        """Return a new object with the same data in a new shape.

        Parameters
        ----------
        *shape
            The new shape as one or more integers or as a tuple.

        Returns
        -------
        obj
            Reshaped object.
        """
        if len(shape) == 1 and isinstance(shape[0], tuple):
            shape = shape[0]
        obj = self.__class__(self.data.reshape(*shape, self.dim))
        obj._data = self._data.reshape(*shape, -1)
        return obj

    def transpose(self, *axes: Optional[int]) -> Object3d:
        """Return a new object with the same data transposed.

        If :attr:`ndim` is 2, the order may be undefined. In this case
        the first two dimensions are transposed.

        Parameters
        ----------
        axes
            The transposed axes order. Only navigation axes need to be
            defined. May be undefined if the object only has two
            navigation dimensions.

        Returns
        -------
        obj
            Transposed object.
        """
        # 1d object should not be transposed
        if len(self.shape) == 1:
            return self

        # allow 2d object to be transposed without specifying axes
        if not len(axes):
            if len(self.shape) != 2:
                raise ValueError("Axes must be defined for more than two dimensions.")
            else:
                # swap first two axes
                axes = (1, 0)

        if len(axes) != len(self.shape):
            raise ValueError(
                "Number of axes is ill-defined: "
                + f"{tuple(axes)} does not fit with {self.shape}."
            )

        return self.__class__(self.data.transpose(*axes + (-1,)))

    def get_random_sample(
        self, size: Optional[int] = 1, replace: bool = False, shuffle: bool = False
    ):
        """Return a new flattened object from a random sample of a given
        size.

        Parameters
        ----------
        size
            Number of samples to draw. Cannot be greater than the size
            of this object. If not given, a single sample is drawn.
        replace
            See :meth:`numpy.random.Generator.choice`.
        shuffle
            See :meth:`numpy.random.Generator.choice`.

        Returns
        -------
        new
            New flattened object of a given size with elements drawn
            randomly from this object.

        See Also
        --------
        numpy.random.Generator.choice
        """
        n = self.size
        if size > n:
            raise ValueError(f"Cannot draw a sample greater than {self.size}")
        rng = np.random.default_rng()
        sample = rng.choice(n, size=size, replace=replace, shuffle=shuffle)
        sample = np.unravel_index(sample, shape=self.shape)
        return self[sample]
